##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


# RQM OSLC API https://jazz.net/wiki/bin/view/Main/RqmOslcQmV2Api

import logging
import re

import anytree
import lxml.etree as ET
import requests
import tqdm

from . import _app
from . import _project
from . import _typesystem
from . import oslcqueryapi
from . import rdfxml
from . import server
from . import utils

#################################################################################################

logger = logging.getLogger(__name__)

#################################################################################################


class _QMProject(_project._Project):
    # A QM project
    def __init__(self, name, project_uri, app, is_optin=False, singlemode=False,defaultinit=True):
        super().__init__(name, project_uri, app, is_optin,singlemode,defaultinit=defaultinit)
        self._components = None  # keyed on component uri
        self._configurations = None # keyed on the config name
        self._folders = None
        self._foldersnotyetloaded = None
        self.is_singlemode = False # this is only true if config enabled is true and single mode is true
        self.gcconfiguri = None
        self.default_query_resource = "oslc_qm:TestCaseQuery"

    def load_components_and_configurations(self,force=False):
        if self._components is not None and self._configurations is not None and not force:
            return
        self._components = {}
        self._configurations = {}
        ncomps = 0
        nconfs = 0
        # retrieve components and configurations for this project
        if not self.is_optin:
            logger.debug( f"{self.is_optin=}" )
            # get the default configuration
            projx = self.execute_get_xml( self.reluri( 'rm-projects/' + self.iid ), intent="Retrieve the project definition (opt-out)" )
            compsu = rdfxml.xmlrdf_get_resource_text( projx, './/jp06:components' )
            compsx = self.execute_get_xml(compsu, intent="Retrieve the component definition (opt-out)" )
            defaultcompu = rdfxml.xmlrdf_get_resource_uri( compsx, './/oslc_config:component' )
            
            # register the only component
            ncomps += 1
            self._components[defaultcompu] = {'name': self.name, 'configurations': {}}
            thisconfu = defaultcompu+"/configurations"
            configs = self.execute_get_json( thisconfu, intent="Retrieve all configurations (opt-out)" )
            configdetails = configs[defaultcompu+"/configurations"]
            if type(configs[thisconfu]["http://www.w3.org/2000/01/rdf-schema#member"])==dict:
                confs = [configs[thisconfu]["http://www.w3.org/2000/01/rdf-schema#member"]]
            else:
                confs = configs[thisconfu]["http://www.w3.org/2000/01/rdf-schema#member"]
            for aconf in confs:
                confu = aconf['value']
                confx = self.execute_get_xml( confu, intent="Retrieve configuration definition (opt-out)" )
                conftitle = rdfxml.xmlrdf_get_resource_text(confx,'.//dcterms:title')
                conftype = 'Stream' if 'stream' in confu else 'Baseline'
                self._components[defaultcompu]['configurations'][confu] = {'name': conftitle, 'conftype': conftype, 'confXml': confx}
                self._configurations[defaultcompu] = self._components[defaultcompu]['configurations'][confu]
                nconfs += 1
        elif self.singlemode:
            logger.debug( f"{self.singlemode=}" )
            #get the single component from a QueryCapability
            # <oslc:QueryCapability>
            #    <oslc_config:component rdf:resource="https://mb02-calm.rtp.raleigh.ibm.com:9443/rm/cm/component/_ln_roBIOEeumc4tx0skHCA"/>
            #    <oslc:resourceType rdf:resource="http://jazz.net/ns/rm/dng/view#View"/>
            #    <oslc:queryBase rdf:resource="https://mb02-calm.rtp.raleigh.ibm.com:9443/rm/views_oslc/query?componentURI=https%3A%2F%2Fmb02-calm.rtp.raleigh.ibm.com%3A9443%2Frm%2Fcm%2Fcomponent%2F_ln_roBIOEeumc4tx0skHCA"/>
            #    <dcterms:title rdf:datatype="http://www.w3.org/2001/XMLSchema#string">View Definition Query Capability</dcterms:title>
            # </oslc:QueryCapability>

            px = self.execute_get_xml( self.project_uri, intent="Retrieve project definition" )

            sx = self.get_services_xml()
            assert sx is not None, "sx is None"
            compuri = rdfxml.xmlrdf_get_resource_uri(sx, ".//oslc:QueryCapability/oslc_config:component")
            assert compuri is not None, "compuri is None"

            ncomps += 1
            self._components[compuri] = {'name': self.name, 'configurations': {}}
            configs = self.execute_get_xml( compuri+"/configurations", intent="Retrieve all project/component configurations (singlemode)" )
            for conf in rdfxml.xml_find_elements(configs,'.//rdfs:member'):
                confu = rdfxml.xmlrdf_get_resource_uri(conf)
                thisconfx = self.execute_get_xml( confu, intent="Retrieve a configuration definition (singlemode)" )
                conftitle= rdfxml.xmlrdf_get_resource_text(thisconfx,'.//dcterms:title')
                # e.g. http://open-services.net/ns/config#Stream
                isstr = rdfxml.xml_find_element( thisconfx,'.//oslc_config:Stream' )
                if isstr is None:
                    conftype = "Baseline"
                else:
                    conftype = "Stream"
                self._components[compuri]['configurations'][confu] = {'name': conftitle, 'conftype': conftype, 'confXml': thisconfx}
                self._configurations[confu] = self._components[compuri]['configurations'][confu]
                nconfs += 1
            self._configurations = self._components[compuri]['configurations']
        else: # full optin
            logger.debug( f"full optin" )
            cmsp_xml = self.app.retrieve_cm_service_provider_xml()
            logger.info( f"cmsp=",ET.tostring(cmsp_xml) )
#  <rdf:Description rdf:nodeID="A4">
#    <oslc:resourceType rdf:resource="http://open-services.net/ns/config#Component"/>
#    <oslc:queryBase rdf:resource="https://jazz.ibm.com:9443/qm/oslc_config/resources/com.ibm.team.vvc.Component"/>
#    <oslc:resourceShape rdf:resource="https://jazz.ibm.com:9443/qm/oslc_config/resourceShapes/com.ibm.team.vvc.Component"/>
#    <dcterms:title rdf:parseType="Literal">Default query capability for Component</dcterms:title>
#    <rdf:type rdf:resource="http://open-services.net/ns/core#QueryCapability"/>
#  </rdf:Description>

            components_uri = rdfxml.xmlrdf_get_resource_uri(cmsp_xml, './/rdf:Description/rdf:type[@rdf:resource="http://open-services.net/ns/core#QueryCapability"]/../oslc:resourceType[@rdf:resource="http://open-services.net/ns/config#Component"]/../oslc:queryBase')
            logger.info( f"{components_uri=}" )
            # get all components
            crx = self.execute_get_xml( components_uri, intent="Retrieve component definition" )

#      <oslc_config:Component rdf:about="https://jazz.ibm.com:9443/qm/oslc_config/resources/com.ibm.team.vvc.Component/_iw4s4EB3Eeus6Zk4qsm_Cw">
#        <dcterms:title rdf:parseType="Literal">SGC Agile</dcterms:title>
#        <oslc:instanceShape rdf:resource="https://jazz.ibm.com:9443/qm/oslc_config/resourceShapes/com.ibm.team.vvc.Component"/>
#        <dcterms:identifier>_iw4s4EB3Eeus6Zk4qsm_Cw</dcterms:identifier>
#        <dcterms:modified rdf:datatype="http://www.w3.org/2001/XMLSchema#dateTime">2020-12-17T14:52:54.318Z</dcterms:modified>
#        <oslc_config:configurations rdf:resource="https://jazz.ibm.com:9443/qm/oslc_config/resources/com.ibm.team.vvc.Component/_iw4s4EB3Eeus6Zk4qsm_Cw/configurations"/>
#        <acc:accessContext rdf:resource="https://jazz.ibm.com:9443/qm/acclist#_rikP0EB1Eeus6Zk4qsm_Cw"/>
#        <process:projectArea rdf:resource="https://jazz.ibm.com:9443/qm/process/project-areas/_rikP0EB1Eeus6Zk4qsm_Cw"/>
#        <oslc:serviceProvider rdf:resource="https://jazz.ibm.com:9443/qm/oslc_config/serviceProviders/configuration"/>
#        <dcterms:relation rdf:resource="https://jazz.ibm.com:9443/qm/service/com.ibm.rqm.integration.service.IIntegrationService/resources/_rikP0EB1Eeus6Zk4qsm_Cw/component/_iw4s4EB3Eeus6Zk4qsm_Cw"/>
#      </oslc_config:Component>

            for component_el in rdfxml.xml_find_elements(crx, f'.//oslc_config:Component/process:projectArea[@rdf:resource="{self.project_uri}"]/..'):
                logger.info( f"{component_el=}" )
                compu = rdfxml.xmlrdf_get_resource_uri(component_el)
                comptitle = rdfxml.xmlrdf_get_resource_text(component_el, './/dcterms:title')
                logger.info( f"Found component {comptitle}" )
                self._components[compu] = {'name': comptitle, 'configurations': {}}
                ncomps += 1
                confu = rdfxml.xmlrdf_get_resource_uri(component_el, './/oslc_config:configurations')
                configs_xml = self.execute_get_rdf_xml( confu, intent="Retrieve all project/component configuration definitions" )
                # Each config:     <ldp:contains rdf:resource="https://jazz.ibm.com:9443/qm/oslc_config/resources/com.ibm.team.vvc.Configuration/_qT1EcEB4Eeus6Zk4qsm_Cw"/>

                for confmemberx in rdfxml.xml_find_elements(configs_xml, './/ldp:contains'):
                    thisconfu = rdfxml.xmlrdf_get_resource_uri( confmemberx )
                    try:
                        thisconfx = self.execute_get_rdf_xml( thisconfu, intent="Retrieve a configuration definition" )
                        conftitle = rdfxml.xmlrdf_get_resource_text(thisconfx, './/dcterms:title')
                        conftype = rdfxml.xmlrdf_get_resource_uri(thisconfx, './/rdf:type')
                        logger.info( f"Found config {conftitle} {conftype} {thisconfu}" )
                        self._components[compu]['configurations'][thisconfu] = {'name': conftitle, 'conftype': conftype,
                                                                                'confXml': thisconfx}
                        self._configurations[thisconfu] = self._components[compu]['configurations'][thisconfu]
                        nconfs += 1
                    except requests.exceptions.HTTPError as e:
                        pass

        # now create the "components"
        for cu, cd in self._components.items():
            cname = cd['name']
            if not self.is_optin:
                c = self
            else:
                c = self._create_component_api(cu, cname)
            c._configurations = self._components[cu]['configurations']
            self._components[cu]['component'] = c
        return (ncomps, nconfs)

    def get_local_config(self, name_or_uri):
        for cu, cd in self._configurations.items():
            logger.debug( f"{cu=} {cd=} {name_or_uri=}" )
            if cu == name_or_uri or cd['name'] == name_or_uri:
                return cu
        return None

    # load the typesystem using the OSLC shape resources
    def _load_types(self,force=False):
        logger.debug( f"load type {self=} {force=}" )

        # if already loaded, try to avoid reloading
        if self.typesystem_loaded and not force:
            return

        self.clear_typesystem()

        if self.local_config:
            # get the configuration-specific services.xml
            sx = self.get_services_xml(force=True,headers={'configuration.Context': self.local_config, 'net.jazz.jfs.owning-context': None})
        else:
            # No config - get the services.xml
            sx = self.get_services_xml(force=True)
        if sx:
            shapes_to_load = rdfxml.xml_find_elements(sx, './/oslc:resourceShape')

            pbar = tqdm.tqdm(initial=0, total=len(shapes_to_load),smoothing=1,unit=" results",desc="Loading ETM shapes")

            for el in shapes_to_load:
                self._load_type_from_resource_shape(el)
                pbar.update(1)

            pbar.close()
        else:
            raise Exception( "services xml not found!" )

        self.typesystem_loaded = True
        return None

    # pick all the attributes from a resource shape definition
    # and for enumerated attributes get all the enumeration values
    def _load_type_from_resource_shape(self, el, supershape=None):
        return self._generic_load_type_from_resource_shape(el, supershape=None)

    # return a dictionary with all local component uri as key and name as value (so two components could have the same name?)
    def get_local_component_details(self):
        results = {}
        for compuri, compdetail in self._components.items():
            results[compuri] = compdetail['name']
        return results

    def find_local_component(self, name_or_uri):
        self.load_components_and_configurations()
        for compuri, compdetail in self._components.items():
            logger.info( f"Checking {name_or_uri} {compdetail}" )
            if compuri == name_or_uri or compdetail['name'] == name_or_uri:
                return compdetail['component']
        return None

    def _create_component_api(self, component_prj_url, component_name):
        logger.info( f"CREATE QM COMPONENT {self=} {component_prj_url=} {component_name=} {self.app=} {self.is_optin=} {self.singlemode=}" )
        result = _QMComponent(component_name, component_prj_url, self.app, self.is_optin, self.singlemode, defaultinit=False, project=self)
        return result


    def is_type_uri(self, uri):
        if uri and uri.startswith(self.app.baseurl) and '/types/' in uri:
            return True
        return False

    # for OSLC query, given a type URI, return its name
    # qm-specific resolution
    def app_resolve_uri_to_name(self, uri):
        if self.is_folder_uri(uri):
            result = self.folder_uritoname_resolver(uri)
        elif self.is_resource_uri(uri):
            result = self.resource_id_from_uri(uri)
        elif self.is_type_uri(uri):
            result = self.type_name_from_uri(uri)
        else:
            result = None
        return result

    # for OSLC query, given a type URI, return the type name
    def type_name_from_uri(self, uri):
        logger.info( f"finding type name {uri}" )
        if self.is_type_uri(uri):
            try:
                # handle artifact formats (these don't have a title or label in the returned XML)
                if match:=re.search("#([a-zA-Z0-9_]+)$",uri ):
                    id = match.group(1)
                else:
                    # retrieve the definition
                    resource_xml = self.execute_get_rdf_xml( reluri=uri, intent="Retrieve type definition to get its name")
                    # check for a rdf label (used for links, maybe other things)
                    id = rdfxml.xmlrdf_get_resource_text(resource_xml,".//rdf:Property/rdfs:label") or rdfxml.xmlrdf_get_resource_text(resource_xml,".//oslc:ResourceShape/dcterms:title") or rdfxml.xmlrdf_get_resource_text(resource_xml,f'.//rdf:Description[@rdf:about="{uri}"]/rdfs:label')
                    if id is None:
                        id = f"STRANGE TYPE {uri}"
                        raise Exception( f"No type for {uri=}" )
            except requests.HTTPError as e:
                if e.response.status_code==404:
                    logger.info( f"Type {uri} doesn't exist!" )
                    raise
                else:
                    raise
            return id
        raise Exception(f"Bad type uri {uri}")

    def is_resource_uri(self, uri):
        if uri and uri.startswith(self.app.baseurl) and '/reQQsources/' in uri:
            return True
        return False

    # for OSLC query, given a resource URI, return the requirement dcterms:identifier
    def resource_id_from_uri(self, uri):
        if self.is_resource_uri(uri):
            resource_xml = self.execute_get_rdf_xml(reluri=uri, intent="Retrieve resource dcterms:identifier")
            id = rdfxml.xmlrdf_get_resource_text(resource_xml, ".//dcterms:identifier")
            return id
        raise Exception(f"Bad resource uri {uri}")

    def is_folder_uri(self, uri):
        if uri and uri.startswith(self.app.baseurl) and '/folders/' in uri:
            return True
        return False

    def folder_nametouri_resolver(self, path_or_uri):
        logger.debug( f"Finding uri {path_or_uri}" )
        if self.is_folder_uri(path_or_uri):
            return path_or_uri
        name = self.load_folders(path_or_uri)
        if name is not None:
            return name
        if path_or_uri in self._folders:
            return self._folders[path_or_uri].folderuri
        raise Exception(f"Folder name {path_or_uri} not found")

    def folder_uritoname_resolver(self,uri):
        logger.debug( f"Finding name {uri}" )
        if not self.is_folder_uri(uri):
            raise Exception( "Folder uri isn't a uri {uri}" )
        thisfolder = self.load_folders(uri)
        if thisfolder is not None:
            return thisfolder.pathname
        logger.info( f"Folder uri {uri} not found")
        return uri

    def _do_find_config_by_name(self, name_or_uri, nowarning=False, include_workspace=True, include_snapshot=True,
                                include_changeset=True):
        if name_or_uri.startswith('http'):
            return name_or_uri
        return self.get_local_config(name_or_uri)

    def get_default_stream_name( self ):
        if self.is_optin and not self.singlemode:
            raise Exception( "Not allowed if compontn is not singlemode!" )
        for configuri,configdetails in self._configurations.items():
            if configdetails['conftype'] == 'Stream':
                return configuri
        raise Exception( "No stream found!" )

#################################################################################################

class _QMComponent(_QMProject):
    def __init__(self, name, project_uri, app, is_optin=False, singlemode=False,defaultinit=True, project=None):
        if not project:
            raise Exception( "You mist provide a project instance when creating a component" )
        super().__init__(name, project_uri, app, is_optin,singlemode,defaultinit=defaultinit)
        self.component_project = project


#################################################################################################

@utils.mixinomatic
class _QMApp(_app._App, oslcqueryapi._OSLCOperations_Mixin, _typesystem.Type_System_Mixin):
    domain = 'qm'
    project_class = _QMProject
    supports_configs = True
    supports_components = True
    supports_reportable_rest = True
    reportable_rest_status = "Application supports Reportable REST but not implemented here yet"
    reportablerestbase='service/com.ibm.rqm.integration.service.IIntegrationService'
    artifactformats = [ # For RR
            '*'
            ,'collections'
            ,'comments'
            ,'comparisons'  # for 7.0.2
            ,'diff'         # for 7.0.2
            ,'linktypes'
            ,'modules'
            ,'processes'
            ,'resources'
            ,'reviews'
            ,'revisions'
            ,'screenflows'
            ,'storyboards'
            ,'terms'
            ,'text'
            ,'uisketches'
            ,'usecasediagrams'
            ,'views'
        ]
    identifier_name = 'Short ID'
    identifier_uri = 'Identifier'

    def __init__(self, server, contextroot, jts=None):
        super().__init__(server, contextroot, jts=jts)
        self.rootservices_xml = self.execute_get_xml(self.reluri('rootservices'), intent="Retrieve QM application rootservices")
        self.serviceproviders = 'oslc_qm_10:qmServiceProviders'
        self.default_query_resource = "oslc_config:Configuration"

        self.version = rdfxml.xmlrdf_get_resource_text(self.rootservices_xml,'.//rqm:version')
        self.majorversion = rdfxml.xmlrdf_get_resource_text(self.rootservices_xml,'.//rqm:majorVersion')
        logger.info( f"Versions {self.majorversion} {self.version}" )

    def _get_headers(self, headers=None):
        result = super()._get_headers()
        result['net.jazz.jfs.owning-context'] = self.baseurl
        if headers:
            result.update(headers)
        return result

    # load the typesystem using the OSLC shape resources listed for all the creation factories and query capabilities
    def load_types(self, force=False):
        self._load_types(force)

    # load the typesystem using the OSLC shape resources
    def _load_types(self,force=False):
        logger.debug( f"load type {self=} {force=}" )

        # if already loaded, try to avoid reloading
        if self.typesystem_loaded and not force:
            return

        self.clear_typesystem()

        # get the services.xml
        sx = self.retrieve_oslc_catalog_xml()
        if sx:
            shapes_to_load = rdfxml.xml_find_elements(sx, './/oslc:resourceShape')

            pbar = tqdm.tqdm(initial=0, total=len(shapes_to_load),smoothing=1,unit=" results",desc="Loading ETM shapes")

            for el in shapes_to_load:
                self._load_type_from_resource_shape(el)
                pbar.update(1)

            pbar.close()
        else:
            raise Exception( "services xml not found!" )

        self.typesystem_loaded = True
        return None

    # given a type URI, return its name
    def resolve_uri_to_name(self, uri, prefer_same_as=True, dontpreferhttprdfrui=True):
        logger.info( f"resolve_uri_to_name {uri=}" )
        if not uri:
            result = None
            return result
        if not uri.startswith('http://') or not uri.startswith('https://'):
        # try to remove prefix
            uri1 = rdfxml.tag_to_uri(uri,noexception=True)
            logger.debug(f"Trying to remove prefix {uri=} {uri1=}")
            if uri1 is None:
                return uri
            if uri1 != uri:
                logger.debug( f"Changed {uri} to {uri1}" )
            else:
                logger.debug( f"NOT Changed {uri} to {uri1}" )
            # use the transformed URI
            uri = uri1
        if not uri.startswith(self.baseurl):
            if self.server.jts.is_user_uri(uri):
                result = self.server.jts.user_uritoname_resolver(uri)
                logger.debug(f"returning user")
                return result
            uri1 = rdfxml.uri_to_prefixed_tag(uri,noexception=True)
            logger.debug(f"No app base URL {self.baseurl=} {uri=} {uri1=}")
            return uri1
        elif not self.is_known_uri(uri):
            if self.server.jts.is_user_uri(uri):
                result = self.server.jts.user_uritoname_resolver(uri)
            else:
                if uri.startswith( "http://" ) or uri.startswith( "https://" ):
                    uri1 = rdfxml.uri_to_prefixed_tag(uri)
                    logger.debug( f"Returning the raw URI {uri} so changed it to prefixed {uri1}" )
                    uri = uri1
                result = uri
            # ensure the result is in the types cache, in case it recurrs the result can be pulled from the cache
            self.register_name(result,uri)
        else:
            result = self.get_uri_name(uri)
        logger.info( f"Result {result=}" )
        return result

    @classmethod
    def add_represt_arguments( cls, subparsers, common_args ):
        '''
        NOTE this is called on the class (i.e. is a class method) because at this point don't know which app with be queried
        '''
        parser_qm = subparsers.add_parser('qm', help='ETM Reportable REST actions', parents=[common_args])
        
        parser_qm.add_argument('artifact_format', choices=cls.artifact_formats, default=None, help=f'CCM artifact format - possible values are {", ".join(cls.artifact_formats)}')

        # SCOPE settings
        parser_qm.add_argument('-p', '--project', default=None, help='Scope: Name of project - required when using module/collection/view/resource/typename ID/typename as a filter')

        parser_qm.add_argument('-r', '--report', action='store_true', help='Report the fields available')

#        # Source Filters - only use one of these at once - all require a project and configuration!
#        rmex1 = parser_qm.add_mutually_exclusive_group()
#        rmex1.add_argument('-c', '--collection', default=None, help='Sub-scope: RM: Name or ID of collection - you need to provide the project and local/global config')
#        rmex1.add_argument('-m', '--module', default=None, help='Sub-scope: RM: Name or ID of module - you need to provide the project and local/global config')
#        rmex1.add_argument('-v', '--view', default=None, help='Sub-scope: RM: Name of view - you need to provide the project and local/global config')
#        rmex1.add_argument('-r', '--resourceIDs', default=None, help='Sub-scope: RM: Comma-separated IDs of resources - you need to provide the project and local/global config')
#        rmex1.add_argument('-t', '--typename', default=None, help='Sub-scope: RM: Name of type - you need to provide the project and local/global config')
        
#        # Output FILTER settings - only use one of these at once
#        parser_qm.add_argument('-a', '--all', action="store_true", help="Filter: Report all resources")
#        parser_qm.add_argument('-d', '--modifiedsince', default=None, help='Filter: only return items modified since this date in format 2021-01-31T12:34:26Z')
        parser_qm.add_argument('-f', '--fields', default=None, help="Filter using xpath")
#        parser_qm.add_argument('-x', '--expandEmbeddedArtifacts', action="store_true", help="Filter: Expand embedded artifacts")
        
#        # various options
#    #    parser_qm.add_argument('--forever', action='store_true', help="TESTING UNFINISHED: save web data forever (used for regression testing against stored data, may not need the target server if no requests fail)" )
#        parser_qm.add_argument('--nresults', default=-1, type=int, help="TESTING UNFINISHED: Number of results expected (used for regression testing against stored data, doesn't need the target server - use -1 to disable checking")
#        parser_qm.add_argument('--pagesize', default=100, type=int, help="Page size for results paging (default 100)")    
        
#        # Output controls - only use one of these at once!
        rmex2 = parser_qm.add_mutually_exclusive_group()
#        rmex2.add_argument('--attributes', default=None, help="Output: Comma separated list of attribute names to report (requires specifying project and configuration)")
        rmex2.add_argument('--schema', action="store_true", help="Output: Report the schema")
#        rmex2.add_argument('--titles', action="store_true", help="Output: Report titles")
#        rmex2.add_argument('--linksOnly', action="store_true", help="Output: Report links only")
#        rmex2.add_argument('--history', action="store_true", help="Output: Report history")
#        rmex2.add_argument('--coverPage', action="store_true", help="Output: Report cover page variables")
#        rmex2.add_argument('--signaturePage', action="store_true", help="Output: Report signature page variables")
#    #    rmex2.add_argument('--size', action="store_true", help="Output: Set size (required for ???)")

    def process_represt_arguments( self, args, allapps ):
        '''
        Process above arguments, returning a dictionayt of parameters to add to the represt base URL
        NOTE this does have some dependency on thje overall 
        
        NOTE this is called on an instance (i.e. not a class method) because by now we know which app is being queried
        '''
        queryparams = {}
        queryurl = ""
        queryheaders={}
        
        if args.schema:
            queryparams['metadata'] = 'schema'
            
        queryurl = self.reluri(self.reportablerestbase) + "/"+ args.artifact_format
            
        if args.report:
            typestodo = []
            # get the schema, walk it building the tree of fields
            schema_x = self.execute_get_xml(queryurl+"?metadata=schema", intent="Retrieve Reportable REST schema").getroot()
#            print( f"{schema_x.tag=}" )
#            print( f"{schema_x=}" )
            el_x = rdfxml.xml_find_element( schema_x, "./xs:element" )
            typestodo=[el_x.get('type')]
            knowntypes={el_x.get('type'):[el_x.get('type')]} # contains path to [parent
            fieldlist = []
            while typestodo:
#                print( f"{knowntypes=}" )
                typetofind = typestodo.pop(0)
                type_x = rdfxml.xml_find_element( schema_x, f'.//xs:complexType[@name="{typetofind}"]' )
#                print( f"Finding {typetofind=} {type_x=}" )
                if type_x is not None:
                    seq_x = rdfxml.xml_find_element( type_x, './xs:sequence' )
                    type_name = type_x.get('name')
                    name_name = type_x.get('type')
    #                print( f"{type_name=}" )
                    if seq_x is not None:
#                        print( f"{typetofind=} {type_name=} {name_name=}" )
                        for subel_x in rdfxml.xml_find_elements(seq_x,'./xs:element'):
                            subeltype = subel_x.get('type')
                            subelname = subel_x.get('name')
    #                        print( f"{subeltype=}" )
    #                        print( f"{subelname=}" )
                            typestr = ""
                            if subeltype.startswith( "xs:"):
#                                print( f"Type {typetofind} Found endpoint {subelname} {subeltype}" )
                                typestr = " "+subeltype                            
                                fieldpath = "/".join(knowntypes[type_name]+[subelname]) + typestr
                                if fieldpath not in fieldlist:
#                                    print( f"Adding {fieldpath}" )
                                    fieldlist.append(fieldpath)
                                
                            elif subeltype not in knowntypes:
                                typestodo.append(subeltype) 
#                                print( f"Type {typetofind} Queued {subelname=} {subeltype=}" )
                                knowntypes[subeltype] = knowntypes[type_name]+[subelname]
    #                        print( f'{"/".join(knowntypes[subeltype]+[subelname])}' )
                                
                    else:
                        raise Exception( f"xs:sequence not found in schema for type {typetofind}" )
            print( "\n".join(sorted(fieldlist) ) )

        if args.fields:
            queryparams['fields'] = args.fields

        return (queryurl,queryparams,queryheaders)



##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


# DN/DNG Reqif export/import
# also can list the defintitions in a project/component
# consider adding ability to create a definition, e.g. by adding modules/artifacts to it

# DNG Reqif API https://jazz.net/wiki/bin/view/Main/DNGReqIF

# reqif factory and query capabilities resource types:
#  http://jazz.net/ns/rm/dng/reqif#ReqIFExport      dng_reqif:ReqIFExport
#  http://jazz.net/ns/rm/dng/reqif#ReqIFPackage     dng_reqif:ReqIFPackage
#  http://jazz.net/ns/rm/dng/reqif#ReqIFImport      dng_reqif:ReqIFImport
#  http://jazz.net/ns/rm/dng/reqif#ReqIFDefinition  dng_reqif:ReqIFDefinition
#

import os
import argparse
import collections
import datetime
import logging
import re
import socket
import time
import webbrowser

import lxml.etree as ET
import requests
import requests_toolbelt

import elmclient.rdfxml as rdfxml
import elmclient.server as server
import elmclient._app as _app
import elmclient.utils as utils

############################################################################

def reqif_main():

    datetimestamp = '{:%Y%m%d-%H%M%S}'.format(datetime.datetime.now())

    # get some defaults which can be overridden in the environment
    JAZZURL     = os.environ.get("QUERY_JAZZURL"    ,"https://jazz.ibm.com:9443" )
    USER        = os.environ.get("QUERY_USER"       ,"ibm" )
    PASSWORD    = os.environ.get("OUERY_PASSWORD"   ,"ibm" )
    JTS         = os.environ.get("QUERY_JTS"        ,"jts" )
    APPSTRINGS  = os.environ.get("QUERY_APPSTRINGS" ,"rm" )
    LOGLEVEL    = os.environ.get("QUERY_LOGLEVEL"   ,"ERROR,DEBUG" )

    allformats = []
    for app in _app._App.__subclasses__():
        if app.artifact_formats:
            allformats.append(f"{app.domain}: "+",".join( app.artifact_formats )+ "." )
    allformats = " ".join(allformats)

    parser = argparse.ArgumentParser(description="Perform Reportable REST query on an application, with results output to CSV and/or XML - use -h to get some basic help")
    parser.add_argument('projectname', help='Name of project')

    # general settings
    parser.add_argument('-A', '--appstrings', default=APPSTRINGS,help=f'Defaults to "rm,jts" - Must be comma-separated list of used domains or domain:contextroot, the FIRST one must be rm. If using nonstandard context roots for just rm like /rrc then specify "rm:rrc,jts" NOTE if jts is not on /jts then e.g. for /myjts use e.g. "rm:rn1,jts:myjts". Default can be set using environment variable QUERY_APPSTRINGS')
    parser.add_argument('-C', '--component', help='The local component (optional, if used you *have* to specify the local configuration using -F)')
    parser.add_argument('-D', '--delaybetween', type=float,default=0.0, help="Delay in seconds between each import/export - use this to reduce overall server load")
    parser.add_argument('-F', '--configuration', default=None, help='Scope: Name of local config - you need to provide the project - defaults to the "Initial Stream" or "Initial Development" +same name as the project')
    parser.add_argument("-J", "--jazzurl", default=JAZZURL, help="jazz server url (without the /jts!) default {JAZZURL} Default can be set using environment variable QUERY_JAZZURL - defaults to https://jazz.ibm.com:9443 which DOESN'T EXIST")
    parser.add_argument('-L', '--loglevel', default=LOGLEVEL,help=f'Set logging on console and (if providing a , and a second level) to file to one of DEBUG, INFO, WARNING, ERROR, CRITICAL, OFF - default is {LOGLEVEL} - can be set by environment variable QUERY_LOGLEVEL')
    parser.add_argument("-P", "--password", default=PASSWORD, help="User password - can be set using env variable OUERY_PASSWORD - set to PROMPT to be prompted at runtime")
    parser.add_argument('-T', '--certs', action="store_true", help="Verify SSL certificates")
    parser.add_argument("-U", "--username", default=USER, help="User id - can be set using environment variable QUERY_USER")
    parser.add_argument('-W', '--cachecontrol', action='count', default=0, help="Used once -W erases cache then continues with caching enabled. Used twice -WW wipes cache and disables caching. Otherwise caching is continued from previous run(s).")
    parser.add_argument('-Z', '--proxyport', default=8888, type=int, help='Port for proxy default is 8888 - used if found to be active - set to 0 to disable')

    # saved credentials
    parser.add_argument('-0', '--savecreds', default=None, help="Save obfuscated credentials file for use with readcreds, then exit - this stores jazzurl, jts, appstring, username and password")
    parser.add_argument('-1', '--readcreds', default=None, help="Read obfuscated credentials from file - completely overrides commandline/environment values for jazzurl, jts, appstring, username and password" )
    parser.add_argument('-2', '--erasecreds', default=None, help="Wipe and delete obfuscated credentials file" )
    parser.add_argument('-3', '--secret', default="N0tSecret-", help="SECRET used to encrypt and decrypt the obfuscated credentials (make this longer for greater security)" )
    parser.add_argument('-4', '--credspassword', action="store_true", help="Prompt user for a password to save/read obfuscated credentials (make this longer for greater security) - NOTE this is by far the best way to secure saved credentials - they're no longer just obfuscated when you use a runtime password!" )

    subparsers = parser.add_subparsers(help='sub-commands',dest='subparser_name')

    parser_create = subparsers.add_parser('create', help='Create a new/update an existing reqif definition in a project/component, adding module(s) or artifact(s)')
    parser_delete = subparsers.add_parser('delete', help='Delete an existing reqif definition in a project/component')
    parser_export = subparsers.add_parser('export', help='Export one or more definitions from a project/component and download the reqifz file')
    parser_import = subparsers.add_parser('import', help='Import one or more reqifz files into a project/component')
    parser_list = subparsers.add_parser('list', help='List the definitions in a project/component')

    parser_export.add_argument('definitionnames',nargs='*',default=[],help='name of export definition- this can be a regex where . matches any character, etc. If you want the regex to match a complete name put ^ at the start and $ at the end - where multiuple names match, these will be exported in alphabetical order')
    parser_export.add_argument('-t', '--timestamp', action='store_true', help='Add a timestamp to all exported file names (both reqif and html)')
    parser_export.add_argument('-O', '--outputdirectory',default=None,help='Output directory for all exported files')

    parser_list.add_argument('definitionnames',nargs='*',default=[],help='name of export definition- this can be a regex where . matches any character, etc. If you want the regex to match a complete name put ^ at the start and $ at the end - where multiuple names match, these will be exported in alphabetical order')
    parser_list.add_argument('-O', '--outputdirectory',default=None,help='Output directory for all exported files')

    parser_import.add_argument( 'ifiles',nargs="*", default=[], help='one or more reqif files or file patterns (e.g. *.reqifz) to import')
    parser_import.add_argument('-I', '--inputdirectory',default='',help='Input directory for all imported files')
    parser_import.add_argument('-O', '--outputdirectory',default=None,help='Output directory for all exported files')

    parser_create.add_argument('definitionname',help='The reqif definition name to create')
    parser_create.add_argument('-a', '--allcores', action="store_true", help="Add all core artifacts (not modules/collections) from the project/component")
    parser_create.add_argument('-f', '--folders', action="store_false", help="Don't include folders in the reqif (defaults to including) - if you need this off, specify it on the last update or on the create")
    parser_create.add_argument('-i', '--identifiers', nargs='*', default=[], help='Use * for all core artifacts or comma-separated list of requirement IDs to add - can specify this option more than once')
    parser_create.add_argument('-l', '--links', action="store_false", help="Don't include links in the reqif (defaults to including) - if you need this off, specify it on the last update or on the create")
    parser_create.add_argument('-m', '--modules', nargs='*', default=[], help='Use * or comma-separated list of module IDs or names of the module to add - for name you can use a regex - can specify this option more than once')
    parser_create.add_argument('-r', '--removeallartifacts', action="store_true", help="When updating, first remove all artifacts/modules/views")
    parser_create.add_argument('-s', '--description', default="-", help="Description for the definition")
    parser_create.add_argument('-t', '--tags', action="store_false", help="Don't include tags in the reqif (defaults to including) - if you need this off, specify it on the last update or on the create")
    parser_create.add_argument('-u', '--update', action="store_true", help="Update the named definition by adding things - it must already exist!")
# NIY    parser_create.add_argument('-p', '--publicviews', nargs='*', default=[], help='* or CSL of public view names')
# NIY    parser_create.add_argument('-v', '--moduleviews', nargs='*', default=[], help='* or CSL of module view names')

    parser_delete.add_argument('definitionnames',nargs='*',default=[],help='One or more names of export definitions to delete - this can be a regex where . matches any character, etc. If you want the regex to match a complete name put ^ at the start and $ at the end')
    parser_delete.add_argument('-n', '--noconfirm', action='store_true', help="Don't prompt to confirm each delete (DANGEROUS!)")

    args = parser.parse_args()

    # setup logging
    levels = [utils.loglevels.get(l,-1) for l in args.loglevel.split(",",1)]
    if len(levels)<2:
        # if only one log level specified, set both the same
        levels.append(None)
    if -1 in levels:
        raise Exception( f'Logging level {args.loglevel} not valid - should be comma-separated one or two values from DEBUG, INFO, WARNING, ERROR, CRITICAL, OFF' )
    utils.setup_logging( filelevel=levels[0], consolelevel=levels[1] )

    logger = logging.getLogger(__name__)

    if args.erasecreds:
        # read the file to work out length
        contentlen = len(open(args.erasecreds,"rb").read())
        # create same-length random data to overwrite
        for i in range(5):
            randomcontent = os.urandom(contentlen)
            open(args.erasecreds,"w+b").write(randomcontent)
        # and delete the file
        os.remove(args.erasecreds)

        print( f"Credentials file {args.erasecreds} overwritten then removed" )
        return 0

    if args.credspassword:
        if args.readcreds is None and args.savecreds is None:
            raise Exception( "When using -4 you must use -0 to specify a file to save credentials into, and/or -1 to specify a credentials file to read" )
        #make sure the user enters at least one character
        credspassword = ""
        while len(credspassword)<1:
            credspassword = getpass.getpass( "Password (>0 chars, longer is more secure)?" )
    else:
        credspassword = "N0tSecretAtAll"

    if args.readcreds:
#        if args.secret is None:
#            raise Exception( "You MUST specify a secret using -3 or --secret if using -0/--readcreads" )
        try:
            args.username,args.password,args.jazzurl,args.appstrings = json.loads( utils.fernet_decrypt(open(args.readcreds,"rb").read(),"=-=".join([socket.getfqdn(),os.path.abspath(args.readcreds),os.getcwd(),getpass.getuser(),args.secret,credspassword])) )
        except (cryptography.exceptions.InvalidSignature,cryptography.fernet.InvalidToken, TypeError):
            raise Exception( f"Unable to decrypt credentials from {args.readcreds}" )
        print( f"Credentials file {args.readcreds} read" )

    if args.savecreds:
        if args.secret is None:
            raise Exception( "You MUST specify a secret using -3 or --secret if using -1/--savecreads" )
        open(args.savecreds,"wb").write(utils.fernet_encrypt(json.dumps([args.username,args.password,args.jazzurl,args.appstrings]).encode(),"=-=".join([socket.getfqdn(),os.path.abspath(args.savecreds),os.getcwd(),getpass.getuser(),args.secret,credspassword]),utils.ITERATIONS))
        print( f"Credentials file {args.savecreds} created" )
        return 0

    # prompt for password
    if args.password is None or args.password=="PROMPT":
        args.password = getpass.getpass(prompt=f'Password for user {args.username}? :')

    ######################################################
    # request proxy config if appropriate
    if args.proxyport != 0:
        server.setupproxy(args.jazzurl,proxyport=args.proxyport)

    ######################################################
    # setup connection to the server and app(s)
    # approots has keys of the domain and values of the context root
    approots = {}
    allapps = {} #keyed by domain
    themainappstring = args.appstrings.split(",")[0]
    themaindomain = server.JazzTeamServer.get_appstring_details(themainappstring)[0]

    if themaindomain != "rm":
        raise Exception( "First appstring must be rm (only rm provides reqif import/export!)" )

    for appstring in args.appstrings.split(","):
        domain,contextroot = server.JazzTeamServer.get_appstring_details(appstring)
        if domain in approots:
            raise Exception( f"Domain {domain} must not appear twice in {args.appstrings}" )
        approots[domain]=contextroot

    # assert the jts default context root if not already specified in args.appstring
    if 'jts' not in approots:
        approots['jts']='jts'

    # create our "server"
    theserver = server.JazzTeamServer(args.jazzurl, args.username, args.password, verifysslcerts=args.certs,appstring=f"jts:{approots['jts']}",cachingcontrol=args.cachecontrol)

    # create all our apps
    for appdom,approot in approots.items():
        allapps[appdom] = theserver.find_app( f"{appdom}:{approot}", ok_to_create=True )

    # get the main app - it's the one we're going to query - it was first in args.appstring
    mainapp = allapps[themaindomain]
    if not mainapp.supports_reportable_rest:
        raise Exception( f"App {themaindomain} {themainappstring} doesn't provide a reportable rest API" )

    ######################################################
    # find the project and if using components fiund the component and configuration
    theproj = mainapp.find_project(args.projectname)

    if theproj is None:
        raise Exception( f"Project '{args.projectname}' not found")

    # assert default for the component name to be the same as the project name
    if args.component is None:
        if theproj.is_optin:
            print( f"Warning - project '{args.projectname}' is opt-in but you didn't specify a component - using default component '{args.projectname}'" )
        args.component = args.projectname

    # not all apps support components, and even if the app does this project may not be opt-in
    if app.supports_components:
        if not theproj.singlemode and not args.component:
            raise Exception( f"Project {args.projectname} supports components so you must provide a component name" )
        if theproj.singlemode:
            args.component = args.projectname
        thecomp = theproj.find_local_component(args.component)
        if not thecomp:
            raise Exception( f"Component '{args.component}' not found in project {args.projectname}" )
        # assert the default configuration for this component if none is specified
        if args.configuration is None:
            args.configuration = thecomp.initial_stream_name()
            print( f"Warning - project '{args.projectname}' is opt-in but for component '{args.component}' you didn't specify a local configuration - using default stream '{thecomp.initial_stream_name()}'" )
        logger.info( f"{args.configuration=}" )
        if theproj.is_optin:
            if args.configuration or theproj.singlemode:
                if theproj.singlemode:
                    if args.configuration is None:
                        # default to the stream
                        args.configuration = thecomp.get_default_stream_name()
                config = thecomp.get_local_config(args.configuration)
                if config is None:
                    raise Exception( f"Configuration '{args.configuration}' not found in component {args.component}" )

                thecomp.set_local_config(config)
                logger.debug( f"LOCAL {config=}" )
            else:
                raise Exception( f"Project {args.projectname} is opt-in so you must provide a local configuration" )
        else:
            if args.configuration is None:
                # default to the stream
                args.configuration = thecomp.get_default_stream_name()
            config = thecomp.get_local_config(args.configuration)
            if config is None:
                raise Exception( f"Configuration '{args.configuration}' not found in component {args.component}" )

        thecomp.set_local_config(config)


        queryon = thecomp
    else:
        queryon = theproj

    ######################################################
    # return True if string contains any of the characters that might be used in a regex
    def isregexp( s ):
        if any([c in s for c in r"^$?*[].+()"]):
            return True
        return False

    ######################################################
    # return a list of dictionaries - one for each entry in definitionnames
    def getmatchingdefs(alldefs,definitionnames):
        results = []
        # look for matches
        matches = {}
        for exportdef in definitionnames:
            if isregexp(exportdef):
                def_re = re.compile( exportdef, re.IGNORECASE )
            else:
                def_re = re.compile( f"^{exportdef}$", re.IGNORECASE )

            for k,v in alldefs.items():
                title = v['dcterms:title']
                if re.search(def_re,title ):
                    matches[k]=dict(v)
                    matches[k]['exportdef']=exportdef
            results.append(dict(matches))
        return results

    if args.subparser_name=='list':
        #################################################################################
        # List the reqif definitions
        # get the reqif definition query URL
        defn_query_u = queryon.get_query_capability_uri("dng_reqif:ReqIFDefinition")
        # query for the definitions
        alldefs = queryon.execute_oslc_query( defn_query_u, select=['*'])
        
#        alldefs = queryon.do_complex_query( "dng_reqif:ReqIFDefinition", select='*' )
        
        logger.debug( f"{alldefs=}" )
        if args.definitionnames:
            rawmatches = getmatchingdefs(alldefs,args.definitionnames )
            # merge all the matches so only get reported once
            matches = {}
            for match in rawmatches:
                matches.update(match)
        else:
            matches = alldefs

        for k in sorted(matches.keys(),key= lambda k: matches[k]['dcterms:title']):
            print( f"{matches[k]['dcterms:title']} {k}" )
            for k1 in sorted(alldefs[k].keys()):
                print( f"    {k1} {alldefs[k][k1]}" )
            print()
                
    elif args.subparser_name=='export':
        #################################################################################
        # Export a reqif definition
        # first find the matching export definitions
        # then export each one sequentially

        # get the reqif definition query URL
        defn_query_u = queryon.get_query_capability_uri("dng_reqif:ReqIFDefinition")
        # query for the definitions
        alldefs = queryon.execute_oslc_query( defn_query_u, select=['*'])
        logger.debug( f"{alldefs=}" )
        # see if specs are specified
        if args.definitionnames:
            rawmatches = getmatchingdefs(alldefs,args.definitionnames )
            # merge all the matches so only get exported once
            matches = {}
            for match in rawmatches:
                matches.update(match)
            matches = [matches]
        else:
            matches = [alldefs]

        # get the export factory
        export_factory_u = queryon.get_factory_uri("dng_reqif:ReqIFExport")

        for match in matches:
            logger.info( f"\n=============================\n{match=}" )
            for reqif_def in match.keys():
                logger.info( f"\n-----------------------------------\n{reqif_def=}" )
                ed = match[reqif_def].get('exportdef','')
                logger.debug( f"Exporting {ed=} {reqif_def} {match=}" )
                content = f"""<rdf:RDF
    xmlns:dcterms="http://purl.org/dc/terms/"
    xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
    xmlns:dng_reqif="http://jazz.net/ns/rm/dng/reqif#">
<dng_reqif:ReqIFExport>
<dng_reqif:definition rdf:resource="{reqif_def}"/>
</dng_reqif:ReqIFExport>
</rdf:RDF>
"""
                logger.debug( f"{content=}" )
                response = queryon.execute_post_rdf_xml( export_factory_u, data=content, cacheable=False, intent="Initiate reqif export" )
                logger.debug( f" {response.status_code=} {response=}" )
                location = response.headers.get('Location')
                if response.status_code == 202 and location is not None:
                    # wait for the tracker to finished
                    result = queryon.wait_for_tracker( location, interval=1.0, progressbar=True, msg=f"Exporting {match[reqif_def]['dcterms:title']}")
                    time.sleep( 1.0 )
                    if result is not None:
                        pkg = rdfxml.xmlrdf_get_resource_uri( result, ".//dcterms:references" )
                        logger.debug( f"{pkg=}" )
                        pkg_x = queryon.execute_get_rdf_xml( pkg, intent="Retrieve reqif export completed tracker result details" )
                        logger.debug( f"{pkg_x=}" )
                        # get the content and its filename
                        content_u =  rdfxml.xmlrdf_get_resource_uri( pkg_x, ".//dng_reqif:content" )
                        response = queryon.execute_get_binary( content_u, intent="Retrieve reqif export completed tracker content" )
                        fname = response.headers.get('Content-Disposition').split( '"', 2 )[1]
                        if fname is None:
                            raise Exception( "No content-disposition!" )
                        if args.timestamp:
                            fname = os.path.splitext(fname)[0]+f".{datetimestamp}.reqifz"
                        if args.outputdirectory:
                            fname = os.path.join( os.getcwd(), args.outputdirectory, fname )
                        print( f"Reqif saved to {fname}" )
                        open( fname, "wb" ).write( response.content )

                        report_u = rdfxml.xmlrdf_get_resource_uri( pkg_x, ".//dng_reqif:report" )
                        response =  queryon.execute_get_binary( report_u, intent="Retrieve reqif export report" )
                        htmlfile = fname+".html"
                        print( f"Report saved to {htmlfile}" )
                        open( htmlfile, "wb" ).write( response.content )

                        # display the report
                        url = f'file://{os.path.abspath(htmlfile)}'
                        webbrowser.open(url, new=2)  # open in new tab

                else:
                    raise Exception( "Odd response to export command, no Location" )

                if args.delaybetween>0:
                    print( "Delaying between exports" )
                    time.sleep(args.delaybetween)

    elif args.subparser_name=='import':
        #################################################################################
        # import a reqifz file
        # get the package factory base URL
        pkg_factory_u = queryon.get_factory_uri("dng_reqif:ReqIFPackage")
        # get the import factory URL
        import_factory_u = queryon.get_factory_uri("dng_reqif:ReqIFImport")
        # accumulate all the input files
        filenames = []
        for ifile in args.ifiles:
            ifile = os.path.join( os.getcwd(), args.inputdirectory, ifile )
            if '*' in ifile or '?' in ifile or '[' in ifile:
                ifiles = glob.glob(ifile)
            else:
                ifiles = [ifile]
            filenames.extend( ifiles )

        for ifile in filenames:
            print( f"Creating upload package for {ifile}" )
            # construct multi-part body containing the file content
            multipart_form_data = collections.OrderedDict(
                    [('upload', (os.path.basename(ifile), open(os.path.normpath(ifile), 'rb'),'application/octet-stream'))
                    ,('userMimeType','application/zip')]
                )
            multipartencoder = requests_toolbelt.MultipartEncoder(multipart_form_data,boundary='---------------------------1336032510420201357832076537')
            content_type, body = multipartencoder.content_type, multipartencoder.to_string()

            # execute post content
            logger.info('Uploading package {ifile}...')
            response = queryon.execute_post_content(pkg_factory_u, data=body, headers={'Content-Type': content_type, 'userMimeType': 'application/zip', 'filename':os.path.basename(ifile),'Accept': '*/*','X-Requested-With': None,'Origin': 'https://jazz.ibm.com:9443', 'OSLC-Core-Version': None}, intent="Initiate reqif export " )

            print( f"Triggering import for {os.path.basename(ifile)}" )

            location = response.headers.get('Location')
            if response.status_code == 201:
                pass
            elif response.status_code == 202 and location is not None:
                # wait for the tracker to finished
                result = queryon.wait_for_tracker( location, interval=1.0, progressbar=True, msg=f"Exporting {match[reqif_def]['dcterms:title']}")
                time.sleep( 1.0 )
                if result is None:
                    raise Exception( f"No result from tracker!" )
            else:
                raise Exception( f"Unknown response {response.status_code}" )

            content=f"""<rdf:RDF
xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
xmlns:dng_reqif="http://jazz.net/ns/rm/dng/reqif#">
    <dng_reqif:ReqIFImport>
         <dng_reqif:package rdf:resource="{location}"/>
   </dng_reqif:ReqIFImport>
</rdf:RDF>"""

            logger.debug( f"{content=}" )
            response = queryon.execute_post_rdf_xml( import_factory_u, data=content, cacheable=False, headers={'net.jazz.jfs.owning-context': queryon.project_uri, 'OSLC-Core-Version': None}, intent="Initiate reqif import" )
            logger.debug( f" {response.status_code=} {response=}" )
            location = response.headers.get('Location')
            if response.status_code == 202 and location is not None:
                # wait for the tracker to finished
                result = queryon.wait_for_tracker( location, interval=1.0, progressbar=True, msg=f"Importing {os.path.basename(ifile)}")
                time.sleep( 1.0 )
                if result is not None:
                    # get the result
                    report_u = rdfxml.xmlrdf_get_resource_uri( result, ".//dcterms:references" )
                    logger.debug( f"{report_u=}" )
                    response =  queryon.execute_get_binary( report_u, intent="Retrieve reqif import result after tracker completed" )
                    htmlfile = ifile+".html"
                    print( f"Report saved to {htmlfile}" )
                    open( htmlfile, "wb" ).write( response.content )

                    # display the report
                    url = f'file://{os.path.abspath(htmlfile)}'
                    webbrowser.open(url, new=2)  # open in new tab
            else:
                raise Exception( "Odd response to export command, no 202" )
            if args.delaybetween>0:
                print( "Delaying between imports" )
                time.sleep(args.delaybetween)

    elif args.subparser_name=='create':
        #################################################################################
        # Create or update a reqif definition
        # if not updating, make sure the definition doesn't already exist
        # if updating make sure it exists
        # then find the things to be added to the definition

        # get the reqif definition query URL
        defn_query_u = queryon.get_query_capability_uri("dng_reqif:ReqIFDefinition")
        defn_factory_u = queryon.get_factory_uri("dng_reqif:ReqIFDefinition")
        # query for the definitions
        alldefs = queryon.execute_oslc_query( defn_query_u, select=['*'])
        logger.debug( f"{alldefs=}" )
        rawmatches = getmatchingdefs(alldefs,[f"^{args.definitionname}$"] )
        logger.debug( f"{rawmatches=}" )
        if args.update:
            # check it exists
            if len(rawmatches[0])==0:
                raise Exception( f"Reqif definition {args.definitionname} must exist" )
            existingdef = rawmatches[0]
            logger.debug( f"{existingdef=}" )
        else:
            # check it doeesn't exist
            if len(rawmatches[0])>0:
                raise Exception( f"Reqif definition {args.definitionname} must not exist" )

            logger.debug( f"{queryon.services_uri=}" )
            logger.debug( f"{queryon=}" )
            if args.update:
                # get the existing definition XML
                existing_u = list(existingdef.keys())[0]
#                logger.debug( f"{existing_u=}" )
                response,etag = queryon.execute_rdf_xml( existing_u, headers={'Accept': 'application/rdf+xml', 'OSLC-Core-Version': None, 'OSLC-Core-Version': None}, return_etag=True, intent="Retrieve project/component existing reqif definition"  )
                defn_x = ET.ElementTree(ET.fromstring(response.content)).getroot()
                etag = response.headers.get('ETag')
                inctags = rdfxml.xmlrdf_get_resource_text( defn_x, './/dng_reqif:includeTags' )
                incfolders = rdfxml.xmlrdf_get_resource_text( defn_x, './/dng_reqif:includeFolders' )
                inclinks = rdfxml.xmlrdf_get_resource_text( defn_x, './/dng_reqif:includeLinks' )
                logger.debug( f"{inctags=}" )
                logger.debug( f"{incfolders=}" )
                logger.debug( f"{inclinks=}" )
                logger.debug( f"{args.removeallartifacts=}" )
                rootdef = rdfxml.xml_find_element( defn_x, './dng_reqif:ReqIFDefinition' )
                if args.removeallartifacts:
                    # delete all artifacts and views in the definitions
                    artincs = rdfxml.xml_find_elements( rootdef, './dng_reqif:include' )
                    logger.debug( f"{artincs=}" )
                    for artinc in artincs:
                        rootdef.remove(artinc)
            else:
                # create
                defn_x = ET.fromstring(f"""<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
           xmlns:dng_reqif="http://jazz.net/ns/rm/dng/reqif#"
           xmlns:acp="http://jazz.net/ns/acp#"
           xmlns:oslc="http://open-services.net/ns/core#"
           xmlns:dcterms="http://purl.org/dc/terms/"
           >
    <dng_reqif:ReqIFDefinition rdf:about="">
            <dcterms:title rdf:datatype="http://www.w3.org/2001/XMLSchema#string">{args.definitionname}</dcterms:title>
            <dng_reqif:includeTags>true</dng_reqif:includeTags>
            <dng_reqif:includeFolders>true</dng_reqif:includeFolders>
            <dng_reqif:includeLinks>true</dng_reqif:includeLinks>

            <dcterms:identifier/>

            <oslc:serviceProvider rdf:resource="{queryon.services_uri}"/>
            <acp:accessControl rdf:resource=""/>
            <dcterms:contributor>https://jazz.ibm.com:9443/jts/users/ibm</dcterms:contributor>
            <dcterms:modified />
            <dng_reqif:relatedReqIFPackages rdf:resource=""/>

    </dng_reqif:ReqIFDefinition>
</rdf:RDF>
""" )
                rootdef = rdfxml.xml_find_element( defn_x, './dng_reqif:ReqIFDefinition' )
                logger.debug( f"{rootdef=}" )
            # for each artifact/module

#            artinc="""
#        <dng_reqif:include>
#              <dng_reqif:ArtifactSelection>
#                    <dng_reqif:artifact rdf:resource="https://jazz.ibm.com:9443/rm/resources/MD__y90-kB6Eeuh3Iiax2L3Ow"/>
#                  </dng_reqif:ArtifactSelection>
#            </dng_reqif:include>
#"""
#                # for each view
#                viewinc = """
#            <dng_reqif:include>
#                  <dng_reqif:ViewDefinitionSelection>
#                        <dng_reqif:usedInModule rdf:resource="https://jazz.ibm.com:9443/rm/resources/MD__y90-kB6Eeuh3Iiax2L3Ow"/>
#                        <dng_reqif:view rdf:resource="https://jazz.ibm.com:9443/rm/views/VW__wvAA0B6Eeuh3Iiax2L3Ow"/>
#                      </dng_reqif:ViewDefinitionSelection>
#                </dng_reqif:include>
#"""

#             <acp:accessControl rdf:resource="https://jazz.ibm.com:9443/rm/accessControl/_-_28tkB6Eeuh3Iiax2L3Ow"/>
#            <dcterms:contributor>https://jazz.ibm.com:9443/jts/users/ibm</dcterms:contributor>
#            <dcterms:modified rdf:datatype="http://www.w3.org/2001/XMLSchema#dateTime">2021-09-03T12:10:25.582Z</dcterms:modified>
#            <dcterms:identifier rdf:datatype="http://www.w3.org/2001/XMLSchema#string">_8500a38e-4382-46d1-b44b-0b8ab4b301c6</dcterms:identifier>
#            <dng_reqif:relatedReqIFPackages rdf:resource="https://jazz.ibm.com:9443/rm/reqif_oslc/definitions/query?reqIFDefinitionUri=https://jazz.ibm.com:9443/rm/reqif/RD_67rw8gyvEeyENOqNxpyqoQ"/>
#            viewincs = []
#            for viewinc in viewincs:
#                # add elements for view
#                pass
            artincs = []
            rmqbase = queryon.get_query_capability_uri("oslc_rm:Requirement")
            # check moduless
            if args.modules:
                # retrieve all modules
                allmodules = queryon.execute_oslc_query(
                        rmqbase
                        ,whereterms=[['rdm_types:ArtifactFormat','=',f'<{rdfxml.RDF_DEFAULT_PREFIX["jazz_rm"]}Module>']]
                        ,select=['*']
                        ,prefixes={rdfxml.RDF_DEFAULT_PREFIX["dcterms"]:'dcterms',rdfxml.RDF_DEFAULT_PREFIX["rdm_types"]:'rdm_types'}
                    )

                logger.debug( f"{len(allmodules)=}" )
                logger.debug( f"{allmodules=}" )
                for k, v in allmodules.items():
                    logger.debug( f"{v.get('dcterms:identifier')} {v.get('dcterms:title')}" )
                # now filter

                if '*' in args.modules:
                    for moduri in allmodules.keys():
                        if moduri not in artincs:
                            artincs.append(moduri)
                else:
                    # choose modules to include
                    allrefs = []
                    for modref in args.modules:
                        logger.debug( f"{modref=}" )
                        if re.match( r"^[\d,]+$", modref ):
                            # if just digits and commas assume a CSL of IDs
                            allrefs.extend( modref.split( "," ) )
                        else:
                            # a name or name regex
                            allrefs.append(modref)
                    # now try to match
                    # build lookup from id to uri
                    ids = {}
                    for k in allmodules:
                        id = allmodules[k].get('dcterms:identifier')
                        if id:
                            ids[id]=k
                    reffound = False
                    for ref in allrefs:
                        logger.debug( f"{ref=}" )
                        if utils.isint(ref):
                            # search for an id
                            if ref in ids:
                                if ids[ref] not in artincs:
                                    artincs.append(ids[ref])
                                    continue
                        else:
                            # search for regex match
                            if re.match(r'^[a-zA-Z0-9 _]+$', ref ):
                                logger.debug( f"nre" )
                                # pure string match
                                # try regex match
                                for k,v in allmodules.items():
                                    if v.get( 'dcterms:title' ) == ref:
                                        if k not in artincs:
                                            artincs.append(k)
                                            continue
                            else:
                                logger.debug( f"re" )
                                # try regex match
                                for k,v in allmodules.items():
                                    if re.search( ref, v.get( 'dcterms:title' ), re.IGNORECASE ):
                                        if k not in artincs:
                                            artincs.append(k)
                                            continue
                        raise Exception( f"No modules found for {ref}!" )
                    # filter for module id/name
                    logger.debug( f"{artincs=}" )

            if args.allcores:
                # retrieve all the core artifacts (note this includes the artifact for each module)
                allartifacts = queryon.execute_oslc_query(
                        rmqbase
#                        ,whereterms=[['rdm_types:ArtifactFormat','=',f'<{rdfxml.RDF_DEFAULT_PREFIX["jazz_rm"]}Module>']]
                        ,select=['dcterms:identifier', 'rm_nav:parent']
                        ,prefixes={rdfxml.RDF_DEFAULT_PREFIX["dcterms"]:'dcterms',rdfxml.RDF_DEFAULT_PREFIX["rm_nav"]:'rm_nav'}
                    )
                # post-filter for non-empty identifier and incldue them
                for k in list(allartifacts.keys()):
                    logger.debug( f"checking {k=}" )
                    if allartifacts[k].get('dcterms:identifier'):
                        if k not in artincs:
                            artincs.append(k)

            if args.identifiers:
                if '*' in args.identifiers:
                    # all core artifacts excluding modules
                    allids=['*']
                    allartifacts = queryon.execute_oslc_query(
                            rmqbase
                            ,whereterms=[['rdm_types:ArtifactFormat','!=',f'<{rdfxml.RDF_DEFAULT_PREFIX["jazz_rm"]}Module>']]
                            ,select=['dcterms:identifier', 'rm_nav:parent']
                            ,prefixes={rdfxml.RDF_DEFAULT_PREFIX["dcterms"]:'dcterms',rdfxml.RDF_DEFAULT_PREFIX["rm_nav"]:'rm_nav'}
                        )
                else:
                    # this is specifically core artifacts, by identifier
                    allids = []
                    for id in args.identifiers:
                        allids.extend(id.split(","))
                    # check all are integer strings:
                    if not all( [utils.isint(s) for s in allids] ):
                        raise Exception( "Must only use integers for -i" )
                    allartifacts = queryon.execute_oslc_query(
                            rmqbase
                            ,whereterms=[['dcterms:identifier' 'in', allids]]
                            ,select=['dcterms:identifier', 'rm_nav:parent']
                            ,prefixes={rdfxml.RDF_DEFAULT_PREFIX["dcterms"]:'dcterms',rdfxml.RDF_DEFAULT_PREFIX["rm_nav"]:'rm_nav'}
                        )
                # build lookup from id to uri
                # find artifacts with an id and a aprent - if the id is being sought add to artincs
                for k in allartifacts:
                    id = allartifacts[k].get('dcterms:identifier')
                    parent = allartifacts[k].get('rm_nav:parent')
                    if id and parent:
                        if id in allids or allids[0]=='*':
                            if k not in artincs:
                                artincs.append(k)
                                logger.debug( f"{id=} {parent=}" )
            logger.debug( f" Found {len(artincs)}" )

            # assemble the XML for the artifact references
            #
            #        <dng_reqif:include>
            #              <dng_reqif:ArtifactSelection>
            #                    <dng_reqif:artifact rdf:resource="https://jazz.ibm.com:9443/rm/resources/MD__y90-kB6Eeuh3Iiax2L3Ow"/>
            #                  </dng_reqif:ArtifactSelection>
            #            </dng_reqif:include>
            for artinc in artincs:
                # add an include
                inc = ET.SubElement( rootdef, rdfxml.uri_to_tag('dng_reqif:include') )
                artsel = ET.SubElement( inc, rdfxml.uri_to_tag('dng_reqif:ArtifactSelection') )
                art = ET.SubElement( artsel, rdfxml.uri_to_tag('dng_reqif:artifact'), {rdfxml.uri_to_tag('rdf:resource'): artinc} )
                logger.debug( "===\n",ET.tostring(inc) )
            rdfxml.xml_find_element( rootdef,'./dng_reqif:includeTags').text = "true" if args.tags else "false"
            rdfxml.xml_find_element( rootdef,'./dng_reqif:includeFolders').text = "true" if args.folders else "false"
            rdfxml.xml_find_element( rootdef,'./dng_reqif:includeLinks').text = "true" if args.links else "false"
            headers = {'OSLC-Core-Version': None}
            if args.update:
                headers.update({ 'if-match': etag } )
                response = queryon.execute_post_rdf_xml( existing_u, data=defn_x, cacheable=False, put=True, headers=headers, intent="Update a project/component reqif definition"  )
            else:
                response = queryon.execute_post_rdf_xml( defn_factory_u, data=defn_x, cacheable=False, headers=headers, intent="Create a project/component reqif definition" )
                logger.debug( f" {response.status_code=} {response=}" )

            location = response.headers.get('Location')


            logger.debug( f"{location=}" )
            print( f"Created!" )
    elif args.subparser_name=='delete':
        #################################################################################
        # Delete a reqif definition
        # need to check the definition exists then delete it
        # get the reqif definition query URL
        defn_query_u = queryon.get_query_capability_uri("dng_reqif:ReqIFDefinition")
        # query for the definitions
        alldefs = queryon.execute_oslc_query( defn_query_u, select=['*'], intent="Retrieve all project/component reqif definitions" )
        logger.debug( f"{alldefs=}" )

        if args.definitionnames:
            rawmatches = getmatchingdefs(alldefs,args.definitionnames )
            # merge all the matches so only get reported once
            matches = {}
            for match in rawmatches:
                matches.update(match)
        else:
            raise Exception( "No definition name provided!" )

        for k in sorted(matches.keys(),key= lambda k: matches[k]['dcterms:title']):
            print( f"Deleting {matches[k]['dcterms:title']}" )
            if not args.noconfirm:
                # get user to confirm by entering Y or press Q to quit
                oktodelete=False
                resp=input("Press y/Y to confirm delete, or q/Q to quit").lower()
                if resp.startswith('y'):
                    oktodelete=True
                elif resp.startswith('q'):
                    break
            else:
                oktodelete=True

            if oktodelete:
                # create DELETE request and send to the definition
                result =  queryon.execute_delete(k, intent="Delete project/component reqif definition" )
                print( f"Deleted {matches[k]['dcterms:title']} {k}" )
            else:
                print( "Not deleted!" )

    else:
        raise Exception( f"Unrecognized command {args.subparser_name}" )

def main():
    runstarttime = time.perf_counter()
    reqif_main()
    elapsedsecs = time.perf_counter() - runstarttime
    print( f"Runtime was {int(elapsedsecs/60)}m {int(elapsedsecs%60):02d}s" )

if __name__ == '__main__':
    main()


##
## Licensed Material - Property of IBM
## (c) Copyright IBM Corp. 2019, 2020, 2021.  All Rights Reserved.
##

# elmclient\examples\represt.py -A rm text -a -C aa.csv -S -L 10 -g 2 -p rm_gc_p1

# DN/DNG Reportable REST API https://jazz.net/wiki/bin/view/Main/DNGReportableRestAPI
#   DN/DNG FAQ on reportable rest https://jazz.net/wiki/bin/view/Main/RRCReportServiceFAQ
# ETM/RQM Reportable REST API https://jazz.net/wiki/bin/view/Main/RqmApi#Resources_and_their_Supported_Op
# EWM/RTC Reportable REST API https://jazz.net/wiki/bin/view/Main/ReportsRESTAPI

import os

import argparse
import csv
import datetime
import getpass
import json
import logging
import socket
import statistics
import time
import traceback
import urllib
import urllib3

import cryptography
import cryptography.fernet
import requests
import lxml.etree as ET

import elmclient.rdfxml as rdfxml
import elmclient.server as server
import elmclient._app as _app
import elmclient.utils as utils

############################################################################

def represt_main():

    # get some defaults which can be overridden in the environment
    JAZZURL     = os.environ.get("QUERY_JAZZURL"    ,"https://jazz.ibm.com:9443" )
    USER        = os.environ.get("QUERY_USER"       ,"ibm" )
    PASSWORD    = os.environ.get("OUERY_PASSWORD"   ,"ibm" )
    JTS         = os.environ.get("QUERY_JTS"        ,"jts" )
    APPSTRINGS  = os.environ.get("QUERY_APPSTRINGS" ,"rm" )
    LOGLEVEL    = os.environ.get("QUERY_LOGLEVEL"   ,"ERROR,DEBUG" )

    allformats = []
    
    #collect all the artifact formats
    for appcls in _app._App.__subclasses__():
        if appcls.artifact_formats:
            allformats.append(f"{appcls.domain}: "+",".join( appcls.artifact_formats )+ "." )
    allformats = " ".join(allformats)
    
    # for arguments common to all sub-parsers
    common_args = argparse.ArgumentParser(description="Perform Reportable REST query on an application, with results output to CSV and/or XML - use -h to get some basic help. NOTE only rm queries are allowed at the moment.", add_help=False)
    
    # the main arguments parser
    parser = argparse.ArgumentParser(description="Perform Reportable REST query on an application, with results output to CSV and/or XML - use -h to get some basic help. NOTE only rm queries are allowed at the moment.")

    # the holder for sub-parsers
    subparsers = parser.add_subparsers(help='sub-commands',dest='subparser_name')

    # general settings which are common acrosss all reportable rest apps
    common_args.add_argument('-A', '--appstrings', default=APPSTRINGS,help=f'Must be comma-separated list of used domains or domain:contextroot, the FIRST one is where the reportable rest query goes, default {APPSTRINGS} If using nonstandard context roots for just rm and gc like /rrc and /thegc then specify "rm:rrc,gc:thegc" NOTE if jts is not on /jts but is on /myjts then add jts: and its context route without leading / e.g. "rm,jts:myjts" to the end of this string. Default can be set using environment variable QUERY_APPSTRINGS')
    common_args.add_argument('-C', '--csvoutputfile', default=None, help='Name of file to save the CSV results to')
    common_args.add_argument('-D', '--delaybetweenpages', type=float,default=0.0, help="Delay in seconds between each page of results - use this to reduce overall server load particularly for large result sets or when retrieving many attributes")
    common_args.add_argument('-E', '--cacheexpiry', type=int, default=7, help="Days to keep cached results from the server (NOTE query results are never cached) - set to 0 to erase current cache and suppress new caching - set to e.g. -7 to erase current cache and then cache for 7 days, set to 7 to maintain the current cache and keep new entries for 7 days")
#    parser.add_argument("-F", "--forcequery", default=None, help="Force use of this exact query - all scope/filter settings are ignored! - this string is added to the reportable rest publish base for your app - e.g. use '/text/*' to do query /publish/text/* - NOTE you need to put the & and ? for any query parameters and include URL escapes in the parameters!")
    common_args.add_argument( '-G', '--pagesize', default=0, type=int, help="Page size for results paging (default is whatever the server does, e.g. 100)")    
    common_args.add_argument('-H', '--forceheader', action='append', default=[], help="Force adding header with value to the query - you must provide the header name=value. NOTE these override headers from the application. If you want to force deleting a header give it the value DELETE. There is no way of forcing a header to have the value DELETE")
    common_args.add_argument("-J", "--jazzurl", default=JAZZURL, help="jazz server url (without the /jts!) default {JAZZURL} Default can be set using environment variable QUERY_JAZZURL")
    common_args.add_argument('-K', '--collapsetags', action="store_true", help="In CSV output rather than naming column for the tag hierarchy, just use the leaf tag name")
    common_args.add_argument('-L','--loglevel', default=LOGLEVEL,help=f'Set logging on console and (if providing a , and a second level) to file to one of DEBUG, INFO, WARNING, ERROR, CRITICAL, OFF - default is {LOGLEVEL} - can be set by environment variable QUERY_LOGLEVEL')
    common_args.add_argument('-M', '--maxresults', default=0, type=int, help="Limit on number of results - may be exceeded by up to one page of results")
    common_args.add_argument('-N', '--again', type=int, default=1, help="Number of times to repeat the REST API call sequence, must be >1 to get statistics fotr REST call duration")
#    parser.add_argument('-N', '--noprogressbar', action="store_false", help="Don't show progress bar during query")
    common_args.add_argument("-P", "--password", default=PASSWORD, help="User password - can be set using env variable OUERY_PASSWORD - set to PROMPT to be prompted at runtime")
    common_args.add_argument('-R', '--forceparameter', action='append', default=[], help="Force adding query name and value to the query URL - you must provide the name=value, the value will be correctly encoded for you. NOTE these override parameters from the application. If you want to force deleting a parameter give it the value DELETE. There is no way of forcing a parameter to have the value DELETE")
    common_args.add_argument('-S', '--sortidentifier', action="store_true", help="If identifier is in results, sort into ascending numeric order of identifier")
    common_args.add_argument('-T', '--certs', action="store_true", help="Verify SSL certificates")
    common_args.add_argument("-U", "--username", default=USER, help="User id - can be set using environment variable QUERY_USER")
    common_args.add_argument('-V', '--verbose', action="store_true", help="Show verbose info during query, including the query URL")
    common_args.add_argument('-W', '--cachecontrol', action='count', default=0, help="Used once -W erases cache then continues with caching enabled. Used twice -WW wipes cache and disables caching. Otherwise caching is continued from previous run(s).")
    common_args.add_argument('-X', '--xmloutputfile', default=None, help='Name of file to save the XML results to')
    common_args.add_argument('-Z', '--proxyport', default=8888, type=int, help='Port for proxy default is 8888 - used if found to be active - set to 0 to disable')


    # various options
#    common_args.add_argument('--nresults', default=-1, type=int, help="TESTING UNFINISHED: Number of results expected (used for regression testing against stored data, doesn't need the target server - use -1 to disable checking")

    # saved credentials
    common_args.add_argument('-0', '--savecreds', default=None, help="Save obfuscated credentials file for use with readcreds, then exit - this stores jazzurl, jts, appstring, username and password")
    common_args.add_argument('-1', '--readcreds', default=None, help="Read obfuscated credentials from file - completely overrides commandline/environment values for jazzurl, jts, appstring, username and password" )
    common_args.add_argument('-2', '--erasecreds', default=None, help="Wipe and delete obfuscated credentials file" )
    common_args.add_argument('-3', '--secret', default="N0tSecret-", help="SECRET used to encrypt and decrypt the obfuscated credentials (make this longer for greater security)" )
    common_args.add_argument('-4', '--credspassword', action="store_true", help="Prompt user for a password to save/read obfuscated credentials (make this longer for greater security)" )

    # add subcommand for each app that supports reportable rest
    for appcls in _app._App.__subclasses__():
        if appcls.supports_reportable_rest:
            appcls.add_represt_arguments( subparsers, common_args )

    args = parser.parse_args()
    
    if not args.appstrings.split(",")[0].startswith(args.subparser_name):
        args.appstrings=f"{args.subparser_name}:{args.subparser_name},{args.appstrings}"
        print( f"{args.subparser_name} added to front of appstrings - using {args.appstrings} if you need a different context root from /{args.subparser_name} use -A to specify it" )
    
    # setup logging
    levels = [utils.loglevels.get(l,-1) for l in args.loglevel.split(",",1)]
    if len(levels)<2:
        # if only one log level specified, set both the same
        levels.append(None)
    if -1 in levels:
        raise Exception( f'Logging level {args.loglevel} not valid - should be comma-separated one or two values from DEBUG, INFO, WARNING, ERROR, CRITICAL, OFF' )
    utils.setup_logging( filelevel=levels[0], consolelevel=levels[1] )
    
    logger = logging.getLogger(__name__)

    if args.erasecreds:
        # read the file to work out length
        contentlen = len(open(args.erasecreds,"rb").read())
        # create same-length random data to overwrite
        for i in range(5):
            randomcontent = os.urandom(contentlen)
            open(args.erasecreds,"w+b").write(randomcontent)
        # and delete the file
        os.remove(args.erasecreds)

        print( f"Credentials file {args.erasecreds} overwritten then removed" )
        return 0

    if args.credspassword:
        if args.readcreds is None and args.savecreds is None:
            raise Exception( "When using -4 you must use -0 to specify a file to save credentials into, and/or -1 to specify a credentials file to read" )
        #make sure the user enters at least one character
        credspassword = ""
        while len(credspassword)<1:
            credspassword = getpass.getpass( "Password (>0 chars, longer is more secure)?" )
    else:
        credspassword = "N0tSecretAtAll"

    if args.readcreds:
#        if args.secret is None:
#            raise Exception( "You MUST specify a secret using -3 or --secret if using -0/--readcreads" )
        try:
            args.username,args.password,args.jazzurl,args.appstrings = json.loads( utils.fernet_decrypt(open(args.readcreds,"rb").read(),"=-=".join([socket.getfqdn(),os.path.abspath(args.readcreds),os.getcwd(),getpass.getuser(),args.secret,credspassword])) )
        except (cryptography.exceptions.InvalidSignature,cryptography.fernet.InvalidToken, TypeError):
            raise Exception( f"Unable to decrypt credentials from {args.readcreds}" )
        print( f"Credentials file {args.readcreds} read" )

    if args.savecreds:
        if args.secret is None:
            raise Exception( "You MUST specify a secret using -3 or --secret if using -1/--savecreads" )
        open(args.savecreds,"wb").write(utils.fernet_encrypt(json.dumps([args.username,args.password,args.jazzurl,args.appstrings]).encode(),"=-=".join([socket.getfqdn(),os.path.abspath(args.savecreds),os.getcwd(),getpass.getuser(),args.secret,credspassword]),utils.ITERATIONS))
        print( f"Credentials file {args.savecreds} created" )
        return 0

    # do a basic check that the target server is in fact running, this way we can give a clear error message
    # to do this we have to get the host and port number from args.jazzurl
    urlparts = urllib.parse.urlsplit(args.jazzurl)
    if ':' in urlparts.netloc:
        serverhost,serverport = urlparts.netloc.rsplit(":",1)
        serverport = int(serverport)
    else:
        serverhost = urlparts.netloc
        if urlparts.scheme=='https':
            serverport=443
        elif urlparts.scheme=='http':
            serverport=80
        else:
            raise Exception( "Unknown scheme in jazzurl {args.jazzurl}" )
    # now try to connect
    if not server.tcp_can_connect_to_url(serverhost, serverport, timeout=2.0):
        raise Exception( f"Server not contactable {args.jazzurl}" )

    # request proxy config if appropriate
    if args.proxyport != 0:
        server.setupproxy(args.jazzurl,proxyport=args.proxyport)

    # approots has keys of the domain and values of the context root
    approots = {}
    allapps = {} #keyed by domain
    themainappstring = args.appstrings.split(",")[0]
    themaindomain = server.JazzTeamServer.get_appstring_details(themainappstring)[0]
#    print( f"MAIN {themainappstring} {themaindomain}" )
    
    for appstring in args.appstrings.split(","):        
        domain,contextroot = server.JazzTeamServer.get_appstring_details(appstring)
#        print( f"Registering {appstring} {domain} {contextroot}" )
        if domain in approots:
            raise Exception( f"Domain {domain} must not appear twice in {args.appstrings}" )
        approots[domain]=contextroot
    # assert the jts default context root if not already specified in args.appstring
    if 'jts' not in approots:
        approots['jts']='jts'
#    print( f"{approots=}" )    

    # create our "server"
    theserver = server.JazzTeamServer(args.jazzurl, args.username, args.password, verifysslcerts=args.certs,appstring=f"jts:{approots['jts']}",cachingcontrol=args.cachecontrol)

    # create all our apps
    for appdom,approot in approots.items():
        allapps[appdom] = theserver.find_app( f"{appdom}:{approot}", ok_to_create=True )
#        print( f"looking for {appdom}:{approot} {allapps[appdom]}" )

    # get the main app - it's the one we're going to query - it was first in args.appstring
    mainapp = allapps[themaindomain]
    if not mainapp.supports_reportable_rest:
        raise Exception( f"App {themaindomain} {themainappstring} doesn't provide a reportable rest API '{mainapp.reportable_rest_status}'" )

    queryurl,queryparams,queryheaders = mainapp.process_represt_arguments( args, allapps )

    if args.pagesize:
        queryparams['size'] = args.pagesize

    if args.forceparameter:
        for p in args.forceparameter:
            pname, pvalue = p.split("=",1)
            if pvalue=="DELETE":
                if pname in queryparams:
                    del queryparams[pname]
            else:
                queryparams[pname] = pvalue

    resultsxmls=[]
    nresults = 0
    
    headers = {
            'OSLC-Core-Version': None # this one seems to be required - it prevents the app from asserting OSLC-Core-Version when making Reportable REST API calls - this seems to affect paged results when the first query required a login :-o
        }
    headers.update(queryheaders) 
    
    if args.forceheader:
        for h in args.forceheader:
            hname, hvalue = h.split("=",1)
            if hvalue=="DELETE":
                if hname in headers:
                    headers[hname] = None
            else:
                headers[hname] = hvalue
    
    # create the full URL with parameters
    # NOTE doing this now allows the next page link to be used unmodified (headers are provided every time)
    parts = list(urllib.parse.urlparse(queryurl))
    parts[4] = urllib.parse.urlencode(queryparams)
    queryurl = urllib.parse.urlunparse(parts)

    if args.verbose:
        print( f"Retrieving {queryurl}" )

    # collect times for each call
    call_durations = []
    
    # save the url for 'again' repeated calls
    theurl = queryurl
    if args.again<1:
        raise Exception( "--again must be >=1")
    for againcounter in range(args.again):
        queryurl = theurl
        npages = 0
        nresults = 0
        while True:
            # retrieve results from this URL
            timer_start = time.perf_counter()
            
            # call the REST API
            result = mainapp.execute_get_xml(reluri=queryurl, headers=headers, cacheable=False, intent="Retrieve Reportable REST content"  )
            
            # calculate and record the duration
            duration = time.perf_counter()-timer_start
            call_durations.append(duration)
            npages += 1
            
            # process the data
            for art in list(result.getroot()):
                print( ".",end='' )
                if againcounter == 0:
                    resultsxmls.append(art)
                    if args.maxresults>0 and len(resultsxmls)>=args.maxresults:
                        logger.info( "Results truncated to {args.maxresults}" )
                        break
                    
            print()
            
            nextlink = result.getroot().get("href",None)
            nresults += int(result.getroot().get(f"{{{rdfxml.RDF_DEFAULT_PREFIX['rrm']}}}totalCount",0))
            if args.maxresults>0 and nresults>=args.maxresults:
                print( f"Hit results limit {args.maxresults=} with {nresults} results" )
                break
                
            if result.getroot().get("rel",'') != "next" or nextlink is None:
                print( "Finished!")
                break
                
            # go round the loop for the next page
            queryurl = nextlink
            
            # if specified by the user, do a delay between pages
            if args.delaybetweenpages > 0.0:
                print( f"Pausing for {args.delaybetweenpages}s" )
                time.sleep(args.delaybetweenpages)
    
    # assemble the complete results into a single XMl file
    results = []
    resultroot= ET.Element('datasource')
    for resultxml in resultsxmls:
        resultroot.append(resultxml)
        
    print( f"Retrieved {len(resultsxmls)} results" )
        
    if args.xmloutputfile is not None:
        open( args.xmloutputfile,"wb").write(ET.tostring(resultroot))

    if args.csvoutputfile is not None:
        headings = []
        rows=[]
        # now walk the XML to produce the output rows
        # go to each result (one row)
#        print( f"{len(list(resultroot))=}" )
        headingsmapping = {}
        for res in list(resultroot):
            row = {}
            # def getcontentrow( node, thisrowdict, allcolumns, level, path, remove_ns=True ):
            row = utils.getcontentrow( res )
            for k in row.keys():
                if args.collapsetags:
                    k1 = k.rsplit("/",1)[1]
                    if k in headingsmapping and headingsmapping[k] != k1:
                        headingsmapping[k1] = k
                else:
                    k1 = k
                if k1 not in headings:
                    headings.append(k1)                
            rows.append(row)
        if not args.collapsetags:
            headings.sort()
        
        if args.sortidentifier:
            if 'identifier' in headings:
                # put identifier in first column!
                headings.remove('identifier')
                headings = ['identifier']+headings
                # and sort the rows by identifier
                rows.sort(key=lambda x: int(x['identifier']))
            else:
                print( "Can't sort - identifier not in headings" )
                
        # write to CSV
        with open(args.csvoutputfile, 'w', newline='') as csvfile:
#            fieldnames = sorted(headings)
            fieldnames = headings
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames, restval='')
            writer.writeheader()
            for row in rows:
                #            print( "k,v",k,v)
                writer.writerow(row)
    else:
        print("results=", results)

    if args.again > 1:
        # print some stats
        if int( len( call_durations ) / npages ) != len( call_durations ) / npages:
            raise( "Odd number of call durations so at least one set of queries was shorter than the others - can't calculate stats!" )
        if npages > 1:
            # multiple pages - combine the times
            print( f"{npages=} {len(call_durations)=} {call_durations=}" )
            samples = [sum(call_durations[i:i+npages]) for i in range(0,len(call_durations),npages)]
            print( f"{samples=}" )
        else:
            print( f"{len(call_durations)=} {call_durations=}" )
            samples = call_durations
        ncalls = len(call_durations)
        call_mean = statistics.mean(samples)/npages
        call_median = statistics.median(samples)/npages
        call_variance = statistics.pvariance(samples)/npages
        print( f"{call_mean=} {call_median=} {call_variance=}" )

def main():
    runstarttime = time.perf_counter()
    represt_main()
    elapsedsecs = time.perf_counter() - runstarttime
    print( f"Runtime was {int(elapsedsecs/60)}m {int(elapsedsecs%60):02d}s" )

if __name__ == '__main__':
    main()


##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


import logging
import urllib

import requests.exceptions

from . import rdfxml
from . import oslcqueryapi
from . import utils
from . import httpops

logger = logging.getLogger(__name__)

#################################################################################################
# a generic jazz application

class _App( httpops.HttpOperations_Mixin ):
    'A generic Jazz application'
    domain = 'UNSPECIFIED APP DOMAIN'
    project_class = None
    artifact_formats = [] # For RR
    reportablerest_baseurl = "publish"
    supports_reportable_rest = False
    reportable_rest_status = "Not supported by application"
    majorVersion = None
    version = None

    def __init__(self, server, contextroot, *, jts=None):
        super().__init__()
        logger.info( f'Creating app {contextroot} {server=}' )
        self.contextroot = contextroot
        self.baseurl = urllib.parse.urljoin(server.baseurl, contextroot) + "/"
        self.jts = jts
        self.server = server
        self.project_areas_xml = None
        self._projects = None
        self.headers = {}
        self.cmServiceProviders = 'oslc_config:cmServiceProviders'
        self.iid=None # app has a dummy (empty) iid
        self.hooks = []
        self.default_query_resource = None

    def retrieve_cm_service_provider_xml(self):
        cm_service_provider_uri = rdfxml.xmlrdf_get_resource_uri(self.rootservices_xml,
                                                                     self.cmServiceProviders)
        rdf = self.execute_get_rdf_xml(cm_service_provider_uri, intent="Retrieve application CM Service Provider" )
        return rdf

    def retrieve_oslc_catalog_xml(self):
        oslccataloguri = rdfxml.xmlrdf_get_resource_uri(self.rootservices_xml, self.serviceproviders)
        if oslccataloguri is None:
            return None
        return self.execute_get_rdf_xml(oslccataloguri, intent="Retrieve application OSLC Catalog (list of projects)")

    # get local headers
    def _get_headers(self, headers=None):
        logger.info( f"app_gh" )
        result = {'X-Requested-With': 'XMLHttpRequest', 'Referer': self.reluri('web'),'OSLC-Core-Version':'2.0'}
        if self.headers:
            result.update(self.headers)
        result.update(self._get_oslc_headers())
        if headers:
            result.update(headers)
        logger.info( f"app_gh {result}" )
        return result

    # get a request with local headers
    def _get_request(self, verb, reluri='', *, params=None, headers=None, data=None):
        fullheaders = self._get_headers()
        if headers is not None:
            fullheaders.update(headers)
        sortedparams = None if params is None else {k:params[k] for k in sorted(params.keys())}
        request = httpops.HttpRequest( self.server._session, verb, self.reluri(reluri), params=sortedparams, headers=fullheaders, data=data)
        return request

    def _get_oslc_headers(self, headers=None):
        result = {
            'Accept': 'application/rdf+xml'
            , 'Referer': self.reluri('web')
            , 'OSLC-Core-Version': '2.0'
        }
        if headers:
            result.update(headers)
        return result

    def find_projectname_from_uri(self,name_or_uri):
        self._load_projects()
            
        if self.is_project_uri(name_or_uri):
            # find the project
            if name_or_uri in self._projects:
                return self._projects[name_or_uri]['name']
        else:
            return name_or_uri

    def is_project_uri(self, uri):
        if uri.startswith(self.baseurl) and '/process/project-areas/' in uri:
            return True
        return False

    # return an absolute URL for a url relative to this app
    # NOTE if reluri has a leading / this will be relative to the serverhostname:port
    # i.e. the app context root will be removed.
    # So if you want an app-relative URL don't use a leading /
    def reluri(self, reluri=''):
        return urllib.parse.urljoin(self.baseurl,reluri)

    # load the projects from the project areas XML - doesn't create any project classes, this is done later when finding a project to open
    def _load_projects(self,include_archived=False,force=False):
        if self.project_class is None:
            raise Exception(f"projectClass has not been set on {self}!")
        if self._projects is not None and not force:
            return
        logger.info( "Loading projects")
        self._projects = {}
        uri = rdfxml.xmlrdf_get_resource_uri(self.rootservices_xml, 'jp06:projectAreas')
        params = {}
        if include_archived:
            params['includeArchived'] = 'true'
        self.project_areas_xml = self.execute_get_xml(uri, params=params, intent="Retrieve all project area definitions" )
        logger.debug( f"{self.project_areas_xml=}" )
        for projectel in rdfxml.xml_find_elements(self.project_areas_xml,".//jp06:project-area" ):
            logger.debug( f"{projectel=}" )
            projectu = rdfxml.xmlrdf_get_resource_text(projectel,".//jp06:url")
            projectname = rdfxml.xmlrdf_get_resource_uri(projectel,attrib='jp06:name')
            logger.debug( f"{projectname=}" )
            is_optin = False
            singlemode = False
            if self.supports_configs:
                en = rdfxml.xmlrdf_get_resource_text(projectel,'.//jp:configuration-management-enabled')
                is_optin = ( rdfxml.xmlrdf_get_resource_text(projectel,'.//jp:configuration-management-enabled') == "true" )
                singlemode = ( rdfxml.xmlrdf_get_resource_text(projectel,'.//jp:configuration-management-mode') == "SINGLE" )
            logger.info( f"{projectname=} {projectu=} {is_optin=} {singlemode=}" )

            self._projects[projectu] = {'name':projectname, 'project': None, 'projectu': projectu, 'is_optin': is_optin, 'singlemode': singlemode }
            self._projects[projectname] = projectu

    # get an instance for a specific project
    def find_project(self, projectname_or_uri, include_archived=False):
        logger.info( f"Find project {projectname_or_uri}")
        self._load_projects()
        if self.is_project_uri(projectname_or_uri):
            if projectname_or_uri in self._projects:
                res = self._projects[projectname_or_uri]['project']
                if res is None:
                    # create the project instance
                    res = self.project_class(self._projects[projectname_or_uri]['name'], self._projects[projectname_or_uri]['projectu'], self, is_optin=self._projects[projectname_or_uri]['is_optin'],singlemode=self._projects[projectname_or_uri]['singlemode'])
            else:
                res = None
        else:
            # must be a name
            projectu = self._projects.get(projectname_or_uri)
            if projectu is None:
                res = None
            else:
                res = self._projects[projectu]['project']
                if res is None:
                    # create the project instance
                    res = self.project_class(self._projects[projectu]['name'], self._projects[projectu]['projectu'], self, is_optin=self._projects[projectu]['is_optin'],singlemode=self._projects[projectu]['singlemode'])
        logger.info( f'Project {projectname_or_uri} found {projectu} {res}' )
        return res

    def report_type_system( self ):
        qcdetails = self.get_query_capability_uris()
        report = "<HTML><BODY>\n"
        report += f"<H1>Type system report for application {self.domain}</H1>\n"
        report += "<H2>Application Queryable Resource Types, short name and URI</H2>\n"
        rows = []
        for k in sorted(qcdetails.keys()):
            shortname = k.split('#')[-1]
            shortname +=  " (default)" if self.default_query_resource is not None and k==rdfxml.tag_to_uri(self.default_query_resource) else ""
            rows.append( [shortname,k,qcdetails[k]])
        # print in a nice table with equal length columns
        report += utils.print_in_html(rows,['Short Name', 'URI', 'Query Capability URI'])
        report += self.textreport()

        rows = []
        for prefix in sorted(rdfxml.RDF_DEFAULT_PREFIX.keys()):
            rows.append([prefix,rdfxml.RDF_DEFAULT_PREFIX[prefix]] )
        report += "<H2>Prefixes</H2>\n"
        report += utils.print_in_html(rows,['Prefix', 'URI'])
        report += "</BODY></HTML>\n"
        return report

    def get_query_capability_uri(self,resource_type=None,context=None):
        context = context or self
        resource_type = resource_type or context.default_query_resource
        return self.get_query_capability_uri_from_xml(capabilitiesxml=context.retrieve_cm_service_provider_xml(), resource_type=resource_type,context=context)

    def get_query_capability_uris(self,resource_type=None,context=None):
        context = context or self
        resource_type = resource_type or context.default_query_resource
        return self.get_query_capability_uris_from_xml(capabilitiesxml=context.retrieve_cm_service_provider_xml(),context=context)

    def get_query_capability_uri_from_xml(self,capabilitiesxml,resource_type,context):
        logger.info( f"get_query_capability_uri_from_xml {self=} {resource_type=} {capabilitiesxml=}" )
        if resource_type is None:
            raise Exception( "You must provide a resource type" )
        # ensure we have a URI for the resource type
        resource_type_u = rdfxml.tag_to_uri(resource_type)
        # get list of [resourcetype,uri]
        qcs = self.get_query_capability_uris_from_xml(capabilitiesxml=capabilitiesxml,context=context)
        if resource_type_u.startswith( 'http' ):
            # looking for a complete precise URI
            if resource_type_u in qcs:
                return qcs[resource_type_u]
            raise Exception( f"Resource type {resource_type} not found" )
        # didn't specify a URI - find the first match at the end of the resouce type
        for k,v in qcs.items():
            if k.endswith(resource_type):
                return v
        raise Exception( f"Query capability {resource_type} {resource_type_u} not found!" )

    # returns a dictionary of resource type to query capability URI
    # this is used when the XML doesn't have references off to other URLs (like GCM does)
    def get_query_capability_uris_from_xml(self,capabilitiesxml,context):
        logger.info( f"get_query_capability_uris_from_xml {self=} {capabilitiesxml=}" )
        qcs = {}
        #<oslc:QueryCapability>
        #    <oslc:resourceType rdf:resource="http://open-services.net/ns/cm#ChangeRequest"/>
        #    <oslc:queryBase rdf:resource="https://jazz.ibm.com:9443/ccm/oslc/contexts/_2H-_4OpoEemSicvc8AFfxQ/workitems"/>
        # find a queryBase and it's the containing tag that has the info
        for qcx in rdfxml.xml_find_elements(capabilitiesxml,'.//oslc:queryBase/..'):
            for qcrtx in rdfxml.xml_find_elements( qcx, 'oslc:resourceType'):
                qcs[rdfxml.xmlrdf_get_resource_uri(qcrtx)] = rdfxml.xmlrdf_get_resource_uri(qcx, "oslc:queryBase")
                logger.debug( f"{rdfxml.xmlrdf_get_resource_uri(qcrtx)=}" )
        return qcs

    def get_factory_uri_from_xml(self,factoriesxml,resource_type,context):
        logger.info( f"get_factory_uri_from_xml {self=} {resource_type=} {factoriesxml=}" )
        if resource_type is None:
            raise Exception( "You must provide a resource type" )
        # ensure we have a URI for the resource type
        resource_type_u = rdfxml.tag_to_uri(resource_type)
        # get list of [resourcetype,uri]
        qcs = self.get_factory_uris_from_xml(factoriesxml=factoriesxml,context=context)
        if resource_type_u.startswith( 'http' ):
            # looking for a complete precise URI
            if resource_type_u in qcs:
                return qcs[resource_type_u]
            raise Exception( f"Factory for resource type {resource_type} not found" )
        # didn't specify a URI - find the first match at the end of the resouce type
        for k,v in qcs.items():
            if k.endswith(resource_type):
                return v
        raise Exception( f"QFactory {resource_type} {resource_type_u} not found!" )

    # returns a dictionary of resource type to factory URI
    # this is used when the XML doesn't have references off to other URLs (like GCM does)
    def get_factory_uris_from_xml(self,factoriesxml,context):
        logger.info( f"get_factory_uris_from_xml {self=} {factoriesxml=}" )
        qcs = {}
        #<oslc:QueryCapability>
        #    <oslc:resourceType rdf:resource="http://open-services.net/ns/cm#ChangeRequest"/>
        #    <oslc:queryBase rdf:resource="https://jazz.ibm.com:9443/ccm/oslc/contexts/_2H-_4OpoEemSicvc8AFfxQ/workitems"/>
        # find a queryBase and it's the containing tag that has the info
        for qcx in rdfxml.xml_find_elements(factoriesxml,'.//oslc:CreationFactory'):
            for qcrtx in rdfxml.xml_find_elements( qcx, 'oslc:resourceType'):
                qcs[rdfxml.xmlrdf_get_resource_uri(qcrtx)] = rdfxml.xmlrdf_get_resource_uri(qcx, "oslc:creation")
                logger.debug( f"{rdfxml.xmlrdf_get_resource_uri(qcrtx)=}" )
        return qcs

    def is_user_uri(self, uri):
        logger.info( f"{self=} {self.jts=}" )
        if uri and uri.startswith(self.jts.baseurl) and '/users/' in uri:
            return True
        return False

    def user_uritoname_resolver(self, uri):
        if self.is_user_uri(uri):
            res = uri[uri.rfind("/") + 1:]
            return res
        raise Exception(f"Bad user uri {uri}")

    def is_user_name(self, name):
        logger.info( f"Checking name {name}" )
        if not name or name.startswith( "http:") or name.startswith( "https:"):
            return False
        res = self.user_nametouri_resolver( name,raiseifinvalid=False)
        if res is not None:
            return True
        return False

    def user_nametouri_resolver(self, name, raiseifinvalid=True):
        logger.info( f"Converting name {name}" )
        if not raiseifinvalid or self.is_user_name(name):
            user_uri = self.jts.baseurl+f"users/{name}"
            # check it using whoami
            try:
                res = self.execute_get(user_uri, intent="Try to retrieve User" )
            except requests.exceptions.HTTPError as e:
                res = None
            if res:
                return user_uri
            else:
                if raiseifinvalid:
                    raise Exception( f"User {name} is not known on this server" )
                return None
        raise Exception(f"Bad user  name {name}")

    def resolve_project_nametouri(self, name, raiseifinvalid=True):
        # find project for name
        self._load_projects()
        result = self._projects.get(name)
        logger.debug( f"resolve_project_nametouri {name} {result}" )
        return result

#################################################################################################

class JTSApp(_App):
    'The JTS application'
    domain = 'jts'
    project_class = None
    supports_configs = False
    supports_components = False
    supports_reportable_rest = False

    def __init__(self, server, contextroot, jts=None):
        super().__init__(server, contextroot, jts=self)


    def find_project(self, projectname):
        raise Exception("JTS does not have projects!")




##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


from .server import *
from ._app import *
from ._project import *
from ._app import *
from ._project import *
from .oslcqueryapi import *
from ._typesystem import *
from ._ccm import *
from ._rm import *
from ._gcm import *
from ._qm import *
from ._relm import *
from .__meta__ import *

__app__ = __meta__.app
__version__ = __meta__.version
__license__ = __meta__.license
__author__ = __meta__.author




##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


import logging
import re

import lark
import lxml.etree as ET

from . import rdfxml

logger = logging.getLogger(__name__)

#
# This enhanced OSLC Query grammar supports "||", "&&" and bracketed evaluation
# This is based on the EBNF in OSLC Query 3.0 standard https://tools.oasis-open.org/version-control/browse/wsvn/oslc-core/trunk/specs/oslc-query.html
# Differences/enhancements from the OSLC 3.0 definition:
#  You can combine compound_terms using && and || - note when using these the compound_term MUST be in ( ) (which also makes reading the expression easier)
#  also accepts just a single compound_term with no ( ) needed (i.e. this is basically the vanilla OSLC Query 3.0 syntax)
#  In order to allow the user to use friendly names for attributes and enumeration values, there is a new terminal valueidentifier - these are resolved in the context of the identifier (attribute) in the LHS of the term - a name with a space must be surrounded by ' '
#  You are of course free to use <https://...> URIs for identifiers and value identifiers. Good luck with that!
#  identifiers and valueidentifiers must start with [a-zA-Z0-9_]
#  Identifiers can't use backslash escapes
#  langtag can be used but will be ignored
#  ^^ type-specifiers can be used and will be passed on into the query, but not checked/enforced
#  In order to allow the user to use names with spaces (e.g. an enumeration value 'In Process'), put ' ' around them
#  Won't resolve identifiers or valueidentifiers if they include a : (can't tell the difference between that usage and e.g. dcterms:identifier)
#  DNG 6.0.6.1 doesn't support the Query 3.0 PrefixedName on RHS of a term as a value, so they are all converted to <https://...> references
#  For DNG You can refer to a foldername using e.g. $"folder name" - this gets converted to the URI of the folder in <> NOT EXTENSIVELY TESTED!
#  You can refer to a user name using e.g. @"user name" - this gets converted to the server-specific URI of the user

_core_oslc_grammar = """
?compound_term      : simple_term
                    | simple_term boolean_op compound_term                    -> do_oslcand

boolean_op          : "and"

?simple_term        : term
                    | scoped_term

?scoped_term        : identifier_wc "{" compound_term "}"

term                : identifier_wc comparison_op value
                    | identifier_wc inop in_val

inop                : "in"

in_val              : "[" invalue ("," invalue)* "]"

invalue             : value
                    | "*" unsignedinteger -> reqid_to_module_uris

?identifier_wc      : identifier | WILDCARD

identifier          : dottedname
                    | prefixedname
                    | simpleidentifier


dottedname          : ( URI_REF_ESC | NAME | "'"SPACYNAME "'" ) "." ( NAME | "'" SPACYNAME "'" )

prefixedname        : ( URI_REF_ESC | NAME ) ":" NAME

simpleidentifier    : ( NAME  | "'" SPACYNAME "'" )

NAME                : /[a-zA-Z0-9_]\w*/

value              : URI_REF_ESC
                    | boolean
                    | decimal
                    | typedliteralstring
                    | literalstring
                    | valueidentifier
                    | urioffoldername
                    | uriofuser
                    | uriofmodule

valueidentifier     : ( ( URI_REF_ESC | NAME | "'" SPACYNAME "'" ) ":" )? NAME
                    | "'" SPACYNAME "'"
                    | "~" unsignedinteger -> reqid_to_core_uri

unsignedinteger     : /[1-9][0-9]*/

URI_REF_ESC         : /<https?:.*?>/

SPACYNAME           : /[a-zA-Z0-9_][^']*/

urioffoldername     : "$" string_esc

uriofuser           : "@" string_esc

uriofmodule         : "^" string_esc

typedliteralstring  : string_esc (langtag | ("^^" prefixedname))

literalstring       : string_esc

boolean             : TRUE | FALSE

WILDCARD            : "*"

comparison_op       : EQ
                    | NE
                    | LT
                    | GT
                    | LE
                    | GE



EQ: "="
NE: "!="
LT: "<"
LE: "<="
GE: ">="
GT: ">"

TRUE            : "true"
FALSE           : "false"

decimal         : SIGNED_NUMBER
newstring_esc      : "'" /[^']+/ "'"
string_esc      : ESCAPED_STRING

langtag         : /@[a-z][a-z0-9_]*/

%import common.ESCAPED_STRING
%import common.SIGNED_NUMBER
%import common.WS
%ignore WS
"""

_enhanced_oslc3_query_grammar = """
where_expression    : compound_term
                    | logicalor_term

?logicalor_term     : logicalcompound_term
                    | logicalcompound_term ( "||" logicalcompound_term )+             -> do_logicalor

?logicalcompound_term    : "(" ( compound_term | logicalor_term ) ")"
                    | "(" ( compound_term | logicalor_term ) ")" ("&&" logicalcompound_term)+                  -> do_logicaland

""" + _core_oslc_grammar

# The basic grammer removes logicalorterm and logicalandterm
_basic_oslc3_query_grammar = """
where_expression    : compound_term

""" + _core_oslc_grammar

# This class will turn a parsed query into a list of steps each (identifier,operation,value) - corresponding to an OSLC query compound_term
# and these lists are combined with the enhanced operations for logicalor and logicaland
# the idea is that results from OSLC query compound_terms are pushed on a stack, and the the logicalor and logicaland take the top
# two entries in the stack and push one result back on the stack.

# The stack will end up with one entry - the results - this is assured by the parsing because it won't generate anything else on a valid enhanced query string

# the transformer does things like turning identifiers (which are human-friendly names) into URIs using nameresolver, and
# turning valueidentifier which are friendly enumeration value names into <URI> (this is done in the term() method)
# while doing this it updates the mapping dictionaries so the results
# can use them to turn result URIs back into friendly names

# leafs of the tree are called first, returning results upwards
class _ParseTreeToOSLCQuery(lark.visitors.Transformer):
    def __init__(self,resolverobject):
        super().__init__()
        self.resolverobject = resolverobject
        self.mapping_uri_to_identifer = {}
        self.mapping_identifer_to_uri = {}
        self.mapping_folders = {} # contains both: key name->uri and key uri->name (uris and names never overlap)
        self.mapping_users = {} # contains both: key name->uri and key uri->name (uris and names never overlap)
        self.mapping_modules = {} # contains both: key name->uri and key uri->name (uris and names never overlap)
        self.mapping_projects = {} # contains both: key name->uri and key uri->name (uris and names never overlap)

    def where_expression(self, s):
        logger.debug( f"where_expression {s=}" )
        result = s
        logger.debug( f"where_expression {s=} returning {result=}" )
        return s

    def do_logicalor(self, s):
        if len(s) == 1:
            result = s[0]
        else:
            if isinstance(s[0][0], str):
                res0 = [s[0]]
            else:
                res0 = s[0]

            if isinstance(s[1][0], str):
                res1 = [s[1]]
            else:
                res1 = s[1]
            result = [res0, res1, "logicalor"]
        return result

    def do_logicaland(self, s):
        if len(s) == 1:
            result = s[0]
        else:
            if isinstance(s[0][0], str):
                res0 = [s[0]]
            else:
                res0 = s[0]

            if isinstance(s[1][0], str):
                res1 = [s[1]]
            else:
                res1 = s[1]
            result = [res0, res1, "logicaland"]
        return result

    def compound_term(self,s):
        return s

    def boolean_op(self,s):
        return "and"

    def do_oslcand(self, s):
        if isinstance(s, list):
            if len(s) == 1:
                result = s
            else:
                result = s
        else:
            raise Exception( f"s isn't a list! {s=}" )
        # put the and first, then term1 then term2
        # check if term2 is another and
        if s[2][0]=='and':
            # if the second term of the and is another and, flatten it out
            result = [s[1],s[0],s[2][1], s[2][2]]
        else:
            # otherwise rearrange to be 'and' term1 term2
            result = [s[1],s[0],s[2]]
        return result

    def term(self, s):
        # if RHS (or for in any of the items in RHS list) is a valueidentifier without a prefix, this uses the lhs to resolve the valueidentifier, i.e. as context for attribute value names
        # or if RHS is an untyped literal then LHS is used to appy ^^type:name to it to make it a typedliteral
        # check if first elem is a property identifier, and if so see if value(s) are identifiers if so resolve them in context of the first identifier (e.g. for enum values)
        logger.info( f"Term {type(s)} {s}" )
        identifier, op, value = s
        if s[0] != '*':
            #
            if op == "in":
                # the "value" is actually a list of values, each of which must be resolved if it is an identifier
                if not isinstance(value, list):
                    value = [value]
                resultlist = []
                for val in value:
                    if isinstance(val, str) and not val.startswith('"') and ':' not in val:
                        # this is an valueidentifier - try to resolve it as an enum in the context of identifier
                        if self.resolverobject.resolve_enum_name_to_uri is not None:
                            result = self.resolverobject.resolve_enum_name_to_uri(val,identifier)
                            if result is None:
                                raise Exception(f"List ref {val} not resolved in context {identifier}")
                            resultlist.append("<" + result + ">")
                    else:
                        resultlist.append(val)
                    s[2] = resultlist
            else:
                t1 = type(value)
                logger.info( f"t1 {value} {t1=}" )
                if isinstance(value, str) and not value.startswith('"') and not value.startswith("'") and ':' not in value and not re.match("\d",value):
                    # this is a valueidentifier - try to resolve it as an enum in the context of identifier
                    if self.resolverobject.resolve_enum_name_to_uri is not None:
                        result = self.resolverobject.resolve_enum_name_to_uri(value, identifier)
                        if result is None:
                            raise Exception(f"Single ref {value} not resolved in context {identifier}")
                        if result.startswith("http:") or result.startswith("https:"):
                            s[2] = "<" + result + ">"
                        else:
                            s[2] = '"'+result+'"'
        logger.info( f"Term returning {s}" )
        return s

    def simpleidentifier(self,s):
        logger.info( f"simpleidentifier {s=}" )
        if len(s) != 1:
            raise Exception( "Bad simpleidentifier" )
        resultname = s[0].value
        # look it up and if necessary store to mapping
        result = self.resolverobject.resolve_property_name_to_uri(resultname)
        if result is None:
            raise Exception("Name resolution for %s failed!" % (resultname))
        else:
            self.mapping_identifer_to_uri[resultname] = result
            self.mapping_uri_to_identifer[result] = resultname
        logger.info( f"simpleidentifier {result=}" )
        return result

    def prefixedname(self,s):
        logger.info( f"prefixedname {s=}" )
        logger.info( f"prefixedname {s=}" )
        result = s[0]+":"+s[1]
        logger.info( f"prefixedname {result=}" )
        return result

    def dottedname(self,s):
        logger.info( f"dottedname {s=} {s[0]=}" )
        if len(s) != 2:
            raise Exception( "Bad dottedname" )

        # s[0][0] is the shape name
        # s[0][1] is the proprty name
        shapename = s[0].value
        propname = s[1].value
        shapeuri = self.resolverobject.resolve_shape_name_to_uri(shapename)
        result = self.resolverobject.resolve_property_name_to_uri(propname,shapeuri )
        self.mapping_identifer_to_uri[f"{shapename}.{propname}"] = result
        self.mapping_uri_to_identifer[result] = f"{shapename}.{propname}"
        logger.info( f"dottedname {s=} {s[0]=} returns {result}" )
        return result

    def identifier(self, s):
        logger.info( f"Identifier {s=}" )
        if len(s) == 1:
            if type(s[0])==str:
                result = s[0]
            else:
                result = s[0].value
        elif len(s) > 1:
            raise Exception( "Bad identifier" )
        logger.info( f"Identifier returning {result=}" )
        return result

    def urioffoldername(self,s):
        logger.info( f"urioffoldername {s=}" )
        name=s[0].strip('"')
        if self.resolverobject.folder_nametouri_resolver is not None:
            uri = self.resolverobject.folder_nametouri_resolver(name)
            if uri is None:
                raise Exception( "Folder name {name} not found!" )
            self.mapping_folders[name]=uri
            self.mapping_folders[uri]=name
            result = "<"+uri.folderuri+">"
        else:
            raise Exception( "This application doesn't support folder names!" )
        return result

    def uriofuser(self,s):
        logger.info( f"uriofuser {s=}" )
        name=s[0].strip('"')
        if self.resolverobject.user_nametouri_resolver is not None:
            uri = self.resolverobject.user_nametouri_resolver(name)
            if uri is None:
                raise Exception( "User name {name} not found!" )
            self.mapping_users[name]=uri
            self.mapping_users[uri]=name
            result = "<"+uri+">"
        else:
            raise Exception( "This application doesn't support users names!" )
        return result

    def uriofmodule(self,s):
        logger.info( f"uriofmodule {s=}" )
        name=s[0].strip('"')
        if self.resolverobject.resolve_modulename_to_uri is not None:
            uri = self.resolverobject.resolve_modulename_to_uri(name)
            if uri is None:
                raise Exception( f"Module name {name} not found!" )
            logger.info( f"uriofmodule {uri=}" )
            self.mapping_modules[name]=uri
            self.mapping_modules[uri]=name
            result = "<"+uri+">"
        else:
            raise Exception( "This application doesn't support module names!" )
        return result

#    def uriofproject(self,s):
#        logger.info( f"uriofproject {s=}" )
#        name=s[0].strip('"')
#        if self.resolverobject.resolve_project_nametouri is not None:
#            uri = self.resolverobject.resolve_project_nametouri(name)
#            if uri is None:
#                raise Exception( f"Project name {name} not found!" )
#            logger.info( f"{uri=}" )
#            self.mapping_projects[name]=uri
#            self.mapping_projects[uri]=name
#            result = "<"+uri+">"
#        else:
#            raise Exception( "This application doesn't support project names!" )
#        return result

    def valueidentifier(self, s):
        logger.info( f"valueidentifier {s=}" )
        if len(s)>2:
            raise Exception( f"s should be no more than two items {s=}" )
        elif len(s)==2:
            # this is a prefix:name - check prefix is in the known prefixes and add it to the list of use prefixes
            resultname = s[0].value+":"+s[1].value
            # for DOORS Next always expand the prefixed name to be an actual URI
            result = "<"+rdfxml.tag_to_uri(resultname)+">"
        else:
            result = s[0].value
        logger.info( f"valueidentifier {s=} returning {result}" )
        return result

    def unsignedinteger(self, s):
        logger.info( f"unsignedinteger {s=}" )
        result = s[0]
        logger.info( f"unsignedinteger {s=} returning {result}" )
        return result

    def reqid_to_core_uri(self, s):
        logger.info( f"reqid_to_core_uri {s=}" )
        if len(s)>2:
            raise Exception( f"s should be no more than two items {s=}" )
        elif len(s)==2:
            # this is a prefix:name - check prefix is in the known prefixes and add it to the list of use prefixes
            resultname = s[0].value+":"+s[1].value
            # for DOORS Next always expand the prefixed name to be an actual URI
            result = "<"+rdfxml.tag_to_uri(resultname)+">"
        else:
            result = s[0].value
        # now look it up - using an OSLC query!
        requri = self.resolverobject.resolve_reqid_to_core_uri( result )
        if requri is None:
            raise Exception( f"ID {result} not found!" )
        result = "<"+requri+">"
        logger.info( f"reqid_to_core_uri {s=} returning {result}" )
        return result

    def reqid_to_module_uris(self, s):
        logger.info( f"reqid_to_module_uris {s=}" )
        if len(s)>2:
            raise Exception( f"s should be no more than two items {s=}" )
        elif len(s)==2:
            # this is a prefix:name - check prefix is in the known prefixes and add it to the list of use prefixes
            resultname = s[0].value+":"+s[1].value
            # for DOORS Next always expand the prefixed name to be an actual URI
            result = "<"+rdfxml.tag_to_uri(resultname)+">"
        else:
            result = s[0].value
        # now look it up - using an OSLC query!
        requris = self.resolverobject.resolve_reqid_to_module_uris( result )
        if requris is None:
            raise Exception( f"ID {result} not found!" )
        result = "<"+">,<".join(requris)+">"
        logger.info( f"reqid_to_module_uris {s=} returning {result}" )
        return result

    def comparison_op(self, s):
        logger.info( f"comparison_op {s=}" )
        return s[0].value

    def value(self, s):
        logger.info( f"value {s=}" )
        result = s[0]
        logger.info( f"value {s=} returning {result}" )
        return result

    def invalue(self, s):
        logger.info( f"invalue {s=}" )
        result = s[0]
        logger.info( f"invalue {s=} returning {result}" )
        return result

    def string_esc(self, s):
        logger.info( f"string_esc {s} returning {s[0].value}" )
#        print( f"{s=}" )
#        burp
        return s[0].value  # string literals include double quotes in the value

    def typedliteralstring(self, s):
        logger.info( f"typedliteralstring {s}" )
        if s[1]=="xsd:datetime":
            if not re.match(r'"\d\d\d\d-\d\d-\d\d(T\d\d:\d\d:\d\d((\.|,)\d\d\d)?(Z|[+-]\d\d:\d\d)?)?"',s[0]):
                raise Exception( f'Datetime {s[0]} not valid - must be "yyyy-mm-dd[Thh:mm:ss[./,fff]?[Z|+/-hh:mm]]"' )
        result = s[0]+"^^"+s[1]
        logger.info( f"typedliteralstring {s} returning {result}" )
        return result  # string literals include double quotes in the value

    def literalstring(self, s):
        # check for a datetime literal and annotate it with the xsd
        if re.match(r'"\d\d\d\d-\d\d-\d\d(T\d\d:\d\d:\d\d((\.|,)\d\d\d)?(Z|[+-]\d\d:\d\d)?)?"',s[0]):
            result = s[0]
            # this works for DN but not EWM   result = s[0]+"^^xsd:datetime"
        else:
            result = s[0]
        logger.info( f"literalstring {s} {s[0]} returning {result}" )
        return result  # string literals include double quotes in the value

    def boolean(self, s):
        if s[0].value=="true":
            return True
        elif s[0].value=="false":
            return False
        else:
            raise Exception( f"Boolean value must be either true or false - f{s[0].value} isn't allowed" )

    def decimal(self, s):
        logger.info( f"decimal {s=}" )
        # try to convert to int first
        try:
            result = int(s[0].value)
        except ValueError:
            # otherwise try to convert to float
            result = float(s[0].value)
        return result

    def in_val(self, s):
        if len(s) == 1:
            result = s[0]
        else:
            result = s
        return result

    def inop(self, s):
        logger.info( f"inop {s=}" )
        return "in"

    def scoped_term(self, s):
        logger.info( f"scoped_term {s=}" )
        return [s[0], "scope", s[1:]]

# from https://tools.oasis-open.org/version-control/svn/oslc-core/trunk/specs/oslc-core.html#selectiveProperties
# with slight tweaks to implement identifier
_select_grammar  ="""
select_terms    : properties
properties      : property ("," property)*
property        : dottedname | identifier | wildcard | nested_prop
dottedname      : NAME "." NAME
nested_prop     : (identifier | wildcard) "{" properties "}"
wildcard        : "*"
identifier      : ( ( URI_REF_ESC | NAME ) ":" )? NAME
URI_REF_ESC     : /<https?:.*>/
NAME            : /[a-zA-Z0-9_]\w*/
"""

_orderby_grammar = """
sort_terms          : sort_term ("," sort_term)*
sort_term           : scoped_sort_terms | signedterm
signedterm          : SIGN ( dottedname | identifier )
dottedname      : NAME "." NAME
scoped_sort_terms   : identifier "{" sort_terms "}"
identifier          : ( ( URI_REF_ESC | NAME ) ":" )? NAME
URI_REF_ESC         : /<https?:.*>/
NAME                : /[a-zA-Z0-9_]\w*/
SIGN                : ( "+" | "-" | ">" | "<" )
"""

# This class will turn a textual orderby specification into a list of orderby terms

# the transformer does things like turning identifiers (which are human-friendly names) into URIs using nameresolver, and
# turning valueidentifier which are friendly enumeration value names into <URI> (this is done in the term() method)

# leafs of the tree are called first, returning results upwards
class _ParseTreeToOSLCOrderBySelect(lark.visitors.Transformer):
#    def __init__(self, shaperesolver=None, nameresolver=None):
    def __init__(self, resolverobject):
        super().__init__()
        self.mapping_uri_to_identifer = {}
        self.mapping_identifer_to_uri = {}
        self.resolverobject = resolverobject
        self.prefixes = {} # prefixes used (will have to be added to oslc.prefix) - NOTE the key is the uri, the value is the prefix!

    def select_terms(self,s):
        return s[0]

    def select_term(self,s):
        return s

    def nested_prop( self,s):
        result = s[0]+"{"+",".join(s[1])+"}"
        return result

    def wildcard(self,s):
        return "*"

    def properties(self,s):
        return s

    def property(self,s):
        return s[0]

    def sort_terms(self,s):
        return s

    def sort_term(self,s):
        return s[0]

    def signedterm(self,s):
        # mapping to always + or -
        signs = { ">": "+", "<": '-', "+": "+", "-": "-"}
        return signs[s[0]]+s[1]

    def scoped_sort_terms(self,s):
        return s

    def identifier(self, s):
        if len(s) == 1:
            resultname = s[0].value
        elif len(s) > 1:
            # a prefixed name
            resultname = ":".join([s[0].value, s[1].value])
            if s[0].value in rdfxml.RDF_DEFAULT_PREFIX:
                self.prefixes[rdfxml.RDF_DEFAULT_PREFIX[s[0].value]]=s[0].value
            else:
                raise Exception( f"Prefix in orderby '{s[0].value}' not found!" )
        # look it up and if necessary store to mapping
        if ":" not in resultname:
            if self.resolverobject.resolve_property_name_to_uri is not None:
                result1 = self.resolverobject.resolve_property_name_to_uri(resultname)
                if result1 is None:
                    raise Exception("Name resolution for %s failed!" % (resultname))
                else:
                    self.mapping_identifer_to_uri[resultname] = result1
                    self.mapping_uri_to_identifer[result1] = resultname
                result = rdfxml.uri_to_prefixed_tag(result1, uri_to_prefix_map=self.prefixes)

            else:
                raise Exception( f"Cannot resolve {resultname} - no name resolver provided! " )
        else:
            # a prefixed name is assumed to be usable directly (the prefix has been added to prefixes)
            result = resultname
        return result

    def dottedname(self,s):
        logger.info( f"dottedname {s=} {s[0]=}" )
        if len(s) != 2:
            raise Exception( "Bad dottedname" )

        # s[0][0] is the shape name
        # s[0][1] is the proprty name
        shapename = s[0].value
        propname = s[1].value
        shapeuri = self.resolverobject.resolve_shape_name_to_uri(shapename)
        result1 = self.resolverobject.resolve_property_name_to_uri(propname,shapeuri )
        result = rdfxml.uri_to_prefixed_tag(result1, uri_to_prefix_map=self.prefixes)
        self.mapping_identifer_to_uri[f"{shapename}.{propname}"] = result
        self.mapping_uri_to_identifer[result] = f"{shapename}.{propname}"
        logger.info( f"dottedname {s=} {s[0]=} returns {result}" )
        return result



##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


import collections
import datetime
import logging
import os.path
import pickle
import requests
import socket
import shutil
import time
import urllib
import urllib3

from . import _app
from . import utils
from . import httpops

logger = logging.getLogger(__name__)

CACHE_FOLDER = '.web_cache'
COOKIE_SAVE_FILE = ".cookies"
WEB_SAVE_FOLDER = "cache"

# The number of days to locally cache responses (can be extended by commandline, or disabled completely)
CACHEDAYS = 7

# this port will be checked for a proxy - if it is there, it will be used for all requests
# (The default proxy port for Telerik Fiddler is 8888)
PROXY_PORT = 8888

# this is the default proxy dictionary for Requests - this must be configured
# by the application code
proxydict = None

# Disable the InsecureRequestWarning so we can quietly control SSL certificate validation
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

##############################################################################################
# CacheControl setup

import cachecontrol as CC
import calendar
import cachecontrol.heuristics
import email.utils
import cachecontrol.caches.file_cache

class _AddDaysHeuristic(cachecontrol.heuristics.BaseHeuristic):
    def __init__(self,days):
        super().__init__()
        self.days = days

    def update_headers(self, response):
        date = email.utils.parsedate(response.headers['date'])
        expires = datetime.datetime(*date[:6]) + datetime.timedelta(days=self.days)
        return {
            'expires' : email.utils.formatdate(calendar.timegm(expires.timetuple())),
            'cache-control' : 'public',
        }

    def warning(self, response):
        msg = 'Automatically cached! Response is Stale.'
        return '110 - "%s"' % msg

##############################################################################################

def caching_save_creds(cachingcontrol):
#    return ( cachingcontrol < 2 )
    return False

def caching_save_data(cachingcontrol):
    return ( cachingcontrol < 2 )

def caching_wipe_cache(cachingcontrol):
    return ( cachingcontrol > 0 )


##############################################################################################

def setupproxy(url,proxyport=PROXY_PORT):
    # If a proxy is running on proxyport, setup proxydict so requests uses the proxy
    global proxydict
    if proxydict is None and proxyport!=0:
        # test if proxy is running
        if tcp_can_connect_to_url('127.0.0.1', proxyport, timeout=2.0):
            # Fiddler is running so setup proxy
            urlparts = urllib.parse.urlsplit(url)
            if urlparts.port:
                proxyurl = urlparts.scheme + "://" + urlparts.netloc[:urlparts.netloc.find(':')]
            else:
                proxyurl = urlparts.scheme + "://" + urlparts.netloc
            # insert the proxy dictionary
            proxydict = {
                            'https':'http://127.0.0.1:'+str(proxyport) # NOTE the proxy URL is "http:" - Fiddler doesn't provide a https proxy by default so don't change this!
                            ,'http':'http://127.0.0.1:'+str(proxyport)
                        }
            logger.info( f'Setting proxy to {proxydict}' )

##############################################################################################

# utility to see if a port is active listening for connections
def tcp_can_connect_to_url(host, port, timeout=5):
    # create an INET, STREAMing socket
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # now connect to the web server on port
    try:
        s.connect((host, port))
        return True
    except OSError:
        return False


#################################################################################################

# the server has two principle purposes:
# 1) to hold the server client connection, login user and password (and some other server-wide settings)
# 2) to hold some apps - at least a JTS - which share these server-wide settings

# this class is a common interface to provide Jazz server-level operations such as managing connections, provide http REST get/post etc.
# and understands how to transparently login the user to Jazz - the application code just makes the request, this manages the login
#  if it's needed

# caching control =0 for full caching, 1 to wipe the cache then use caching, 2 to wipe cache and disable caching

class JazzTeamServer( httpops.HttpOperations_Mixin ):
    def __init__(self, serverhostport, user, password, jtsappstring='jts', verifysslcerts=True, appstring=None, cachingcontrol=0, cachefolder=CACHE_FOLDER):
        logger.info( f"Creating server {appstring=} {jtsappstring=} {verifysslcerts=} {cachingcontrol=}" )
        self.verifysslcerts = verifysslcerts
        self.username = user
        self.password = password
        self.baseurl = serverhostport
        self.__user = user
        self.__password = password
        self.jts = None
        self.auto_retry = True
        self.cachingcontrol = cachingcontrol # 0=caching, 1=wipe cache then cache, 2= no caching
        self.headers = None
        self.cachefolder = cachefolder
        self.apps = []
        self._session = None

        # setup the session
        self._session = JazzTeamServer.__get_client(user, password,cachingcontrol=cachingcontrol, cachefolder=self.cachefolder)
        self._session.verify = verifysslcerts
        self._session.auto_retry = self.auto_retry
        self._session.cachingcontrol = self.cachingcontrol # 0=caching, 1=wipe cache then cache, 2= no caching

        if not hasattr(self._session,'is_authenticated'):
            self._session.is_authenticated = False

        # create any requested apps
        if jtsappstring:
            self.jts = self.add_app(jtsappstring)
            if appstring is not None:
                if type(appstring)==str:
                    if "," in appstring:
                        appstring = appstring.split(",")
                    else:
                        appstring=[appstring]
                elif type(appstring)!=list:
                    raise Exception( "appstring must be either a string of one or multiple comma-separated appstrings, or a list of single appstrings" )
                for app in appstring:
                    self.add_app(app)
        else:
            if appstring:
                raise Exception( f"You can't add app '{appstring}' because you haven't specified the jts appstring" )


    def get_user_password(self, url=None):
        return (self.__user, self.__password)

    def find_app(self, appstring, ok_to_create=False):
        domain,contextroot = self.get_appstring_details(appstring)
        for app in self.apps:
            if app.domain == domain and app.contextroot==contextroot:
                return app
        if ok_to_create:
            return self.add_app(appstring)
        return None

    # app factory
    def add_app(self, appstring):
        logger.info( f"Adding app {appstring=}" )
        domain,contextroot = self.get_appstring_details(appstring)
        # this is a factory for the different apps
        for app in _app._App.__subclasses__():
            if app.domain == domain:
                thisapp = app(self, contextroot=contextroot, jts=self.jts)
                self.apps.append(thisapp)
                return thisapp
        raise Exception(f"Domain '{domain}' not recognized")

#    # used to locate the app for the uri
#    # this is tentative code to allow e.g. links from ccm to rm in query results from ccm to be looked up in rm to get the requirement ID
#    # HOWEVER there's a design problem to solve; how to get the config (maybe require a GC)
#    # AND (much more difficult) how to get a project-like context (when we don't know the project/component) to provide headers
#    def find_app_for_uri( self, uri ):
#        print( f"fafu {uri=}" )
#        if not uri.startswith( self.reluri() ):
#            return None
#        for app in self.apps:
#            print( f"{app.reluri()=}" )
#            if uri.startswith( app.reluri() ):
#                print( f"Found {uri=} {app=}" )
#                return app
#        return None

    # return the absolute URI for a relative URI
    def reluri(self, reluri=''):
        return urllib.parse.urljoin(self.baseurl,reluri)

    # a note about 'appstring':
    #
    # This code is intended to handle general rm/ccm/qm/etc. domains and to be able to handle con-default context roots like /rm23
    # so introduced the idea of an appstring
    #
    # An appstring allows specifying for an app what domain it is (i.e. strictly rm,qm,ccm,gc, etc.), and it's relative uri e.g. /rm1
    # with sensible defaults so it's easy to specify rm on the default /rm by just specifying appstring "rm"
    #
    # appstring is "domain[:(reluri|partialuri)]"
    # so "rm"=>domain rm, context root /rm
    # rm:rm2=>domain rm, context root /rm2
    # ALSO "rm1"=>domain rm, context root /rm1 (domain rm is deduced by taking the first part of the string which is an allowed domain)
    #
    # If you really really wanted to access rm as e.g. /qm1, you'll have to specify rm:qm1
    #
    @staticmethod
    def get_appstring_details(appstring):
        ''' returns (domain,contextroot)'''
        # get the domains of the classes that inherit from _App
        allowed_domains = [a.domain for a in _app._App.__subclasses__()]
        res = appstring.split(":", 1)
        if len(res) < 2:
            if res[0] not in allowed_domains:
                # check for the start of the string being an allowed domain to allow e.g. rm1 is in domain rm
                for dom in allowed_domains:
                    if res[0].startswith(dom):
                        # copy e.g. rm1 as the context root /rm1
                        res.append(res[0])
                        # assert the domain
                        res[0] = dom
            else:
                # the domain is the context root
                res.append(res[0])
        if res[0] not in allowed_domains:
            raise Exception( f"Domain {res[0]} is not in the allowed domains {str(allowed_domains)}" )
        return res

    '''The interface to a Jazz server - handles all HTTP requests'''
    __shared_client_cache = collections.OrderedDict()

    @staticmethod
    def __get_client(username, password, ignorecache=False,cachingcontrol=0, cachefolder=CACHE_FOLDER):
        '''Get shared client session (one using same user/password)'''
        key = (username, password)
        result = JazzTeamServer.__shared_client_cache.get(key)

        if result is not None and ignorecache:
            # force removal of the existing client and starting a new one
            del JazzTeamServer.__shared_client_cache[key]
            result = None

        cacheexpiry = CACHEDAYS

        if result is None:
            # create a new session

            if caching_wipe_cache(cachingcontrol):
                # uncached data
                # if there is an existing data cache remove it
                if os.path.isdir(os.path.join(cachefolder,WEB_SAVE_FOLDER)):
                    logger.info( f"Erasing existing cache" )
                    shutil.rmtree(os.path.join(cachefolder,WEB_SAVE_FOLDER))
                    time.sleep(1.0)

            if caching_save_data(cachingcontrol):
                # cached - create folder for cache
                webcachefolder = os.path.join(cachefolder,WEB_SAVE_FOLDER)
                os.makedirs(webcachefolder,exist_ok=True)
                # cache to file with the CC heuristic to make responses persist for a number of days
                result = CC.CacheControl(requests.Session(), heuristic=_AddDaysHeuristic(cacheexpiry), cache=CC.caches.file_cache.FileCache(webcachefolder))
                # restore cookies saved after previous login, perhaps we'll avoid having to re-login
            else:
                # use an ordinary session
                result = requests.session()

            if caching_save_creds(cachingcontrol):
                # if credentials are being cached, load them from previous session
                # from https://stackoverflow.com/questions/13030095/how-to-save-requests-python-cookies-to-a-file
                os.makedirs(cachefolder,exist_ok=True)
                if os.path.isfile(os.path.join(cachefolder,COOKIE_SAVE_FILE)):
                    with open(os.path.join(cachefolder,COOKIE_SAVE_FILE), 'rb') as f:
                        result.cookies.update(pickle.load(f))
            else:
                # remove any saved cookies from previous login
                if os.path.isfile(os.path.join(cachefolder,COOKIE_SAVE_FILE)):
                    os.remove(os.path.join(cachefolder,COOKIE_SAVE_FILE))

            JazzTeamServer.__shared_client_cache[key] = result

        # ensure proxies are setup
        result.proxies = proxydict
        result.username = username
        result.password = password
        return result

    @staticmethod
    def clear_client_cache():
        JazzTeamServer.__shared_client_cache = collections.OrderedDict()

    # get local headers
    def _get_headers(self, headers=None):
        logger.info( f"server_gh" )
        result = {}
        if self.headers is not None:
            result.update(self.headers)
        if headers is not None:
            result.update(headers)
        logger.info( f"server_gh {result}" )
        return result

#    # get a request with local headers
    def _get_request(self, verb, reluri='', *, params=None, headers=None, data=None):
        fullheaders = self._get_headers()
        if headers is not None:
            fullheaders.update(headers)
        sortedparams = None if params is None else {k:params[k] for k in sorted(params.keys())}
        request = httpops.HttpRequest( self.app.server._session, verb, self.reluri(reluri), params=sortedparams, headers=fullheaders, data=data)
        return request



